from numpy import sign
from simses.commons.config.simulation.system import StorageSystemConfig
from simses.logic.thermal_management.on_off_controller import OnOffController
from simses.logic.thermal_management.pid_controller import PIDController
from simses.system.auxiliary.heating_ventilation_air_conditioning.hvac import HeatingVentilationAirConditioning
from simses.system.auxiliary.heating_ventilation_air_conditioning.fan import Fan


class FixCOPHeatingVentilationAirConditioning(HeatingVentilationAirConditioning):

    def __init__(self, hvac_configuration: list):
        super().__init__()

        # Get optional user-defined values required for this class
        try:
            self.__max_thermal_power: float = float(hvac_configuration[StorageSystemConfig.HVAC_POWER])
        except IndexError:
            raise Exception('Please specify thermal power for HVAC class' + self.__name__)
        self.__min_thermal_power: float = self.__max_thermal_power * 0.0  # W
        try:
            self.__set_point_temperature = float(hvac_configuration[StorageSystemConfig.HVAC_TEMPERATURE_SETPOINT]) + 273.15  # K
        except (IndexError, ValueError):
            self.__set_point_temperature = 298.15  # K, default value 25 °C
        # source for scop and seer :
        # https://data.toshiba-klima.at/de/Multisplit%20R32%20-%2010,00%20kW%20-%20R32%20-%20Home%20RAS-5M34U2AVG-E%20de.pdf
        # seasonal coefficient of performance (for cooling)
        self.__scop: float = 4.08
        # seasonal energy efficiency ratio (for heating)
        self.__seer: float = 6.31

        # Initialize variables
        self.__electric_power: float = 0  # W
        self.__thermal_power: float = 0  # W
        self.__air_mass: float = 0  # kg
        self.__air_specific_heat: float = 0  # J/kgK

        # Create the Fan
        self.__fan = Fan()

        # Create Thermal Management
        # self.__thermal_management = OnOffController(self.__set_point_temperature, self.__max_thermal_power)
        self.__thermal_management = PIDController(self.__set_point_temperature, self.__max_thermal_power)

    def run_air_conditioning(self, temperature_time_series: [float], temperature_timestep: float, ambient_air_temperature: float) -> None:
        # This HVAC model monitors the air temperature (with respect to the provided set-point)
        # This logic reduces chatter and rapid switching between heating and cooling to some extent
        self.__thermal_power = 0  # W
        self.__electric_power = 0  # W

        total_thermal_power, hvac_thermal_power = self.__thermal_management.compute(temperature_time_series[0], ambient_air_temperature, self.__air_mass, self.__air_density, temperature_timestep, self.__fan)
        self.__thermal_power = float(total_thermal_power)  # W
        hvac_thermal_power = float(hvac_thermal_power)

        # thermal_power is +ve when cooling and -ve when heating
        if abs(total_thermal_power) > 0 and hvac_thermal_power == 0:
            self.__electric_power = self.__fan.electricity_consumption  # W
        elif hvac_thermal_power < 0:
            self.__electric_power = abs(hvac_thermal_power / self.__seer) + self.__fan.electricity_consumption  # W
        elif hvac_thermal_power > 0:
            self.__electric_power = abs(hvac_thermal_power / self.__scop) + self.__fan.electricity_consumption  # W
        else:
            self.__electric_power = 0.0  # W

    def get_max_thermal_power(self) -> float:
        return self.__max_thermal_power

    def get_thermal_power(self) -> float:
        return self.__thermal_power

    def get_electric_power(self) -> float:
        return self.__electric_power

    def get_set_point_temperature(self) -> float:
        return self.__set_point_temperature

    def get_scop(self) -> float:
        return self.__scop

    def get_seer(self) -> float:
        return self.__seer

    def get_temperature_dead_band(self) -> float:
        return self.__thermal_management.TEMPERATURE_DEAD_BAND

    def update_air_parameters(self, air_mass: float = None, air_specific_heat: float = None, air_density: float = None) -> None:
        self.__air_mass = air_mass
        self.__air_specific_heat = air_specific_heat
        self.__air_density = air_density

    def set_electric_power(self, electric_power: float) -> None:
        """Used in cases where multiple runs of the HVAC take place within a SimSES timestep"""
        self.__electric_power = electric_power



from math import log10
import pandas as pd
import scipy.interpolate

from simses.commons.config.data.redox_flow import RedoxFlowDataConfig
from simses.commons.config.simulation.redox_flow import RedoxFlowConfig
from simses.commons.log import Logger
from simses.commons.state.technology.redox_flow import RedoxFlowState
from simses.technology.redox_flow.stack.electrolyte.abstract_electrolyte import ElectrolyteSystem
from simses.technology.redox_flow.stack.abstract_stack import StackModule


class HighPerformanceStack5700W(StackModule):
    """HighPerformanceStack5700W describes a stack based on laboratory measurement cell data at ZAE Bayern from Tobias
    Greese with high-performance electrodes in an interdigitated flow design. The data is scaled up to a 20-cell stack
    with 1000 cm^2 cell area.
    The internal resistance is dependent of the current density. The dependency from SOC and flow rate is neglected. """

    """The nominal power is calculated for an SOC range from 20 % to 80 % and the listed parameters. It corresponds to 
    the power that can be obtained over this SOC range."""
    __STACK_POWER_NOM = 5700  # W
    __CELL_NUMBER = 12  # -
    __CELL_AREA = 1000  # cm^2
    __ELECTRODE_THICKNESS = 0.033  # cm; thin carbon papers
    __ELECTRODE_POROSITY = 0.85  # -
    __MIN_CELL_VOLTAGE = 1.2  # V
    __MAX_CELL_VOLTAGE = 1.6  # V
    __HYDRAULIC_RESISTANCE = 3.082E10  # 1/m^3
    __DEPENDENT_PARAMETERS = True

    def __init__(self, electrolyte_system: ElectrolyteSystem, voltage: float, power: float,
                 redox_flow_data_config: RedoxFlowDataConfig, redox_flow_config: RedoxFlowConfig):
        super().__init__(electrolyte_system, voltage, power, self.__CELL_NUMBER, self.__STACK_POWER_NOM,
                         redox_flow_config)
        self.__log: Logger = Logger(__name__)
        self.__electrolyte_system: ElectrolyteSystem = electrolyte_system
        self.__RINT_FILE = redox_flow_data_config.rfb_rint_file_hp_stack
        self.__SD_FILE = redox_flow_data_config.high_performance_stack_self_discharge
        self.__SHUNT_FILE = redox_flow_data_config.high_performance_stack_shunt
        self.__internal_resistance = pd.read_csv(self.__RINT_FILE, header=None)  # Ohm
        self.__i_rint_arr = self.__internal_resistance.iloc[0, :]
        self.__rint = self.__internal_resistance.iloc[1, :]
        self.__rint_inter1d = scipy.interpolate.interp1d(self.__i_rint_arr, self.__rint, kind='linear',
                                                         fill_value='extrapolate')
        self.__self_discharge_current = pd.read_csv(self.__SD_FILE, header=None)  # mA/cm^2
        self.__i_stack_sd_arr = self.__self_discharge_current.iloc[0, :]
        self.__i_sd_arr = self.__self_discharge_current.iloc[1, :]
        self.__i_sd_inter1d = scipy.interpolate.interp1d(self.__i_stack_sd_arr, self.__i_sd_arr, kind='linear',
                                                         fill_value='extrapolate')
        self.__shunt_current = pd.read_csv(self.__SHUNT_FILE, header=None)  # A
        self.__shunt_current_mat = self.__shunt_current.iloc[1:, 1:]
        self.__soc_arr = self.__shunt_current.iloc[1:, 0]
        self.__current_arr = self.__shunt_current.iloc[0, 1:]
        self.__shunt_interp2d = scipy.interpolate.interp2d(self.__current_arr, self.__soc_arr, self.__shunt_current_mat,
                                                           kind='linear')

    def get_open_circuit_voltage(self, redox_flow_state: RedoxFlowState) -> float:
        """
        Literature source: Fink, Holger. Untersuchung von Verlustmechanismen in Vanadium-Flussbatterien. Diss.
        Technische Universität München, 2019.
        equation 5.18, assumption: SOH = 100 %, therefore ver = 0.5
        """
        concentration_v = self.__electrolyte_system.get_vanadium_concentration()
        soc_stack = redox_flow_state.soc_stack
        # equation is defined for SOC = ]0-1[
        if soc_stack > 0.999:
            soc_stack = 0.999
        if soc_stack < 0.001:
            soc_stack = 0.001
        temperature = redox_flow_state.temperature
        concentration_h_start = 2.6  # mol/l
        ocv_cell = (1.255 + 0.07 + 0.059 * temperature / 298.15 * log10((soc_stack / (1 - soc_stack) *
                    (concentration_h_start + concentration_v / 1000 * (soc_stack + 0.5)))**2 * (concentration_h_start +
                    concentration_v / 1000 * (soc_stack - 0.5))))
        self.__log.debug('OCV cell: ' + str(ocv_cell) + ' V')
        return ocv_cell * self.get_cell_per_stack() * self.get_serial_scale()

    def get_nominal_voltage_cell(self) -> float:
        """Calculated for a temperature of 25 °C and at SOC 50 %."""
        nominal_voltage_cell = 1.423
        return nominal_voltage_cell

    def get_internal_resistance(self, redox_flow_state: RedoxFlowState) -> float:
        current_density = redox_flow_state.current / self.get_specific_cell_area() / self.get_parallel_scale() * 1000
        # mA/cm^2
        resistance = self.__rint_inter1d(current_density)
        self.__log.debug('Resistance stack: ' + str(resistance) + ' Ohm, resistance charging module:'
                         + str(resistance / self.get_parallel_scale() * self.get_serial_scale()))
        return float(resistance / self.get_parallel_scale() * self.get_serial_scale())

    def get_cell_per_stack(self) -> int:
        return self.__CELL_NUMBER

    def get_min_voltage(self) -> float:
        return self.__MIN_CELL_VOLTAGE * self.get_cell_per_stack() * self.get_serial_scale()

    def get_max_voltage(self) -> float:
        return self.__MAX_CELL_VOLTAGE * self.get_cell_per_stack() * self.get_serial_scale()

    def get_self_discharge_current(self, redox_flow_state: RedoxFlowState) -> float:
        current_density = abs(redox_flow_state.current / self.get_specific_cell_area() / self.get_parallel_scale() *
                              1000)  # mA/cm^2
        self_discharge_membrane = self.__i_sd_inter1d(current_density) / 1000 * self.get_specific_cell_area()
        current_stack = redox_flow_state.current / self.get_parallel_scale()
        self_discharge_shunt = (float(self.__shunt_interp2d(redox_flow_state.soc_stack, current_stack)))
        total_self_discharge_current = self_discharge_membrane + self_discharge_shunt
        return (total_self_discharge_current * self.get_cell_per_stack() * self.get_serial_scale() *
                self.get_parallel_scale())

    def get_specific_cell_area(self) -> float:
        return self.__CELL_AREA

    def get_electrode_thickness(self) -> float:
        return self.__ELECTRODE_THICKNESS / 100

    def get_electrode_porosity(self) -> float:
        return self.__ELECTRODE_POROSITY

    def get_hydraulic_resistance(self) -> float:
        return self.__HYDRAULIC_RESISTANCE / (self.get_serial_scale() * self.get_parallel_scale())

    def dependent_parameters(self) -> bool:
        return self.__DEPENDENT_PARAMETERS

    def close(self):
        super().close()
        self.__log.close()
        self.__electrolyte_system.close()



import math

from simses.commons.constants import Hydrogen


class AlkalineFluidModel:   # TODO: more detailed approach with losses instead of just Faraday

    def __init__(self):
        self.__koh_weight_concentration = 30

    def __h2_generation_cell(self, current_cell) -> float:
        return current_cell / (2 * Hydrogen.FARADAY_CONST)  # mol/s

    def __o2_generation_cell(self, current_cell) -> float:
        return current_cell / (4 * Hydrogen.FARADAY_CONST)  # mol/s

    def __h2o_use_cell(self, current_cell) -> float:
        return current_cell / (2 * Hydrogen.FARADAY_CONST)  # mol/s

    def __molarity(self, stack_temperature) -> float:
        # Calculate molarity of KOH electrolyte
        # (From "A review of specific conductivities of potassium hydroxide solution for various concentrations and temperatures" by R.J. Gilliam et al)
        molarity = (self.__koh_weight_concentration * (
                183.1221 - 0.56845 * stack_temperature + 984.5679 * math.exp(self.__koh_weight_concentration / 115.96277))) / (
                           100 * 56.105)
        return molarity

    def get_h2_generation_cell(self, current_cell) -> float:
        return self.__h2_generation_cell(current_cell)

    def get_o2_generation_cell(self, current_cell) -> float:
        return self.__o2_generation_cell(current_cell)

    def get_h2o_use_cell(self, current_cell) -> float:
        return self.__h2o_use_cell(current_cell)

    def get_molarity(self, stack_temperature) -> float:
        return self.__molarity(stack_temperature)



from configparser import ConfigParser


class SystemParameters:
    """
    SystemParameters collects all static information of the storage system and writes it to the results folder.
    """

    SECTION: str = 'System'
    EXTENSION: str = '.txt'

    ID: str = 'id'
    SYSTEM: str = 'system'
    SUBSYSTEM: str = 'subsystems'
    PARAMETERS: str = 'parameters'

    # StorageSystemAC level
    AUXILIARIES: str = 'auxiliaries'
    POWER_DISTRIBUTION: str = 'power_distribution'
    CONTAINER_NUMBER: str = 'number_of_containers'
    CONTAINER_TYPE: str = 'container_type'
    ACDC_CONVERTER: str = 'acdc_converter'

    # StorageSystemDC level
    DCDC_CONVERTER: str = 'dcdc_converter'
    STORAGE_TECHNOLOGY: str = 'technology'
    BATTERIES: str = 'batteries'
    BATTERY_CIRCUIT: str = 'battery_circuit'
    CELL_TYPE: str = 'cell_type'
    NOMINAL_VOLTAGE: str = 'nominal_voltage'

    def __init__(self):
        self.__parameters: ConfigParser = ConfigParser()
        self.__parameters.add_section(self.SECTION)

    def set(self, parameter: str, value: str) -> None:
        """
        Setting a parameters value

        Parameters
        ----------
        parameter :
            Parameter provided by SystemParameter class
        value :
            value to be written to parameter as string

        Returns
        -------

        """
        self.__parameters.set(self.SECTION, parameter, value)

    def set_all(self, parameters: dict) -> None:
        """
        Setting all parameters to internal system parameters

        Parameters
        ----------
        parameters :
            a set of parameters written to system parameters

        Returns
        -------

        """
        for parameter, value in parameters.items():
            self.set(parameter, str(value))

    def write_parameters_to(self, path: str) -> None:
        """
        Writing system parameters to file in path

        Parameters
        ----------
        path :
            path of file to write parameters

        Returns
        -------

        """
        with open(path + self.get_file_name(), 'w') as file:
            self.__parameters.write(file)

    @classmethod
    def get_file_name(cls) -> str:
        """

        Returns
        -------
        str:
            file name of system parameters
        """
        return cls.__name__ + cls.EXTENSION



import pandas as pd

from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.electric.default import DefaultElectricalCellProperties
from simses.technology.lithium_ion.cell.electric.properties import ElectricalCellProperties
from simses.technology.lithium_ion.cell.format.abstract import CellFormat
from simses.technology.lithium_ion.cell.format.round import RoundCell
from simses.technology.lithium_ion.cell.thermal.default import DefaultThermalCellProperties
from simses.technology.lithium_ion.cell.thermal.properties import ThermalCellProperties
from simses.technology.lithium_ion.cell.type import CellType


class GenericCell(CellType):
    """An GenericCell is a special cell type and inherited by CellType"""

    __INTERNAL_RESISTANCE: float = 0.010  # Ohm

    __DIAMETER: float = 22  # mm
    __LENGTH: float = 70  # mm

    __ELECTRICAL_PROPS: ElectricalCellProperties = DefaultElectricalCellProperties()
    __THERMAL_PROPS: ThermalCellProperties = DefaultThermalCellProperties()
    __CELL_FORMAT: CellFormat = RoundCell(__DIAMETER, __LENGTH)

    def __init__(self, voltage: float, capacity: float, soh: float, battery_config: BatteryConfig):
        super().__init__(voltage, capacity, soh, self.__ELECTRICAL_PROPS, self.__THERMAL_PROPS, self.__CELL_FORMAT,
                         battery_config)
        self.__log: Logger = Logger(type(self).__name__)

    def get_open_circuit_voltage(self, battery_state: LithiumIonState) -> float:
        return (3.0 + battery_state.soc) * self.get_serial_scale()

    def get_internal_resistance(self, battery_state: LithiumIonState) -> float:
        return float(self.__INTERNAL_RESISTANCE / self.get_parallel_scale() * self.get_serial_scale())

    def close(self):
        self.__log.close()



from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.technology.hydrogen.electrolyzer.degradation.cyclic.cyclic_degradation import \
    CyclicDegradationModel


class CyclicDegradationPemElMultiDimAnalyticPtlCoating(CyclicDegradationModel):

    """ Anode Ti-PTL of Electrolyzer cell is coated with thin Pt. This avoids the cyclic degradation of the
     resistance. Therefore this degradation model does not have cyclic degradation. """

    def __init__(self):
        super().__init__()
        self.__resistance_increase = 0
        self.__exchange_current_dens_decrease = 0  # p.u.
        self.__current_zero_counter = 0  # counts timesteps since begin of current = 0

    def calculate_resistance_increase(self, time: float, state: ElectrolyzerState) -> None:
       """" No cyclic resistance increase """
       self.__resistance_increase = 0

    def calculate_exchange_current_dens_decrerase(self, state: ElectrolyzerState):
        """ No cyclic exchange_current_density_decrease """
        self.__exchange_current_dens_decrease = 0

    def get_resistance_increase(self) -> float:
        return self.__resistance_increase

    def get_exchange_current_dens_decrease(self) -> float:
        return self.__exchange_current_dens_decrease

    def reset(self, state: ElectrolyzerState) -> None:
        pass

    def close(self) -> None:
        pass






from simses.commons.config.analysis.market import MarketProfileConfig
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.profile.economic.market import MarketProfile
from simses.commons.profile.file import FileProfile
from simses.commons.timeseries.interpolation.last_value import LastValue


class IntradayMarketProfile(MarketProfile):
    """
    Provides the intraday market prices
    """
    def __init__(self, general_config: GeneralSimulationConfig, config: MarketProfileConfig):
        super().__init__()
        self.__file: FileProfile = FileProfile(general_config, config.intraday_price_file, interpolation=LastValue())

    def next(self, time: float) -> float:
        return self.__file.next(time)

    def initialize_profile(self):
        return self.__file.initialize_file()

    def close(self):
        self.__file.close()



from configparser import ConfigParser

import pandas
import pytest

from simses.analysis.data.system import SystemData
from simses.analysis.evaluation.technical.technical_evaluation import *
from simses.commons.config.analysis.general import GeneralAnalysisConfig
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.state.system import SystemState

from simses.analysis.evaluation.technical.system import SystemTechnicalEvaluation


def create_analysis_config():
    analysis_config: ConfigParser = ConfigParser()
    return GeneralAnalysisConfig(config=analysis_config)

timestep = 1
n = 2
yearstart = 2020

def create_general_config() -> GeneralSimulationConfig:
    simulation_config: ConfigParser = ConfigParser()
    simulation_config.add_section('GENERAL')
    simulation_config.set('GENERAL', 'TIME_STEP', str(timestep))

    return GeneralSimulationConfig(simulation_config)

# build configs for testing
gen_sim_config = create_general_config()
analysis_config = create_analysis_config()

dstart = datetime(yearstart, 1, 1).timestamp()
time_test = pandas.array([dstart + timestep * i for i in range(n)])

# Round-trip efficiency:
@pytest.mark.parametrize('storage_power, soc, result',
                          [
                              ([0, -3600 * 10],     [0.1, 0.09],    100),   # only one-way discharged, 100%
                              ([0, 3600*10],        [0, 0.005],     50),    # only one-way charged, 50%
                              ([3600*10, -3600 * 10], [0, 0],      100),   # round trip, 100%, no soc-change
                              ([3600*10, -1800*10], [0, 0],      50),      # round-trip, 50%, no soc-change
                              ([3600 * 10, -1800 * 10], [0, 0.005], 100),  # round trip, 100%, soc-change
                              ([3600 * 10, -1800 * 10], [0, 0.0025], 71.076758)  # round-trip, 50%, soc-change
                          ]
                         )
def test_round_trip_efficiency(storage_power, soc, result):

    capacity = [1000, 1000]
    system_dict = {SystemState.TIME: time_test, SystemState.AC_POWER_DELIVERED: storage_power,
                   SystemState.SOC: soc, SystemState.CAPACITY: capacity,
                   SystemState.SYSTEM_AC_ID: 1, SystemState.SYSTEM_DC_ID: 1}
    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))

    uut = TechnicalEvaluation(system_data, analysis_config)
    assert abs(uut.round_trip_efficiency - result) <= 1e-5


# Capacity_remaining:
@pytest.mark.parametrize('soh, result',
                          [
                              ([0, 0], 0),
                              ([0, 0.5], 50),
                              ([0, -1], -100),
                              ([0, 2], 200)
                          ]
                         )
def test_capacity_remaining(soh, result):

    system_dict = {SystemState.TIME: time_test, SystemState.SOH: soh,
                   SystemState.SYSTEM_AC_ID: 1, SystemState.SYSTEM_DC_ID: 1}
    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))

    uut = TechnicalEvaluation(system_data, analysis_config)
    assert abs(uut.capacity_remaining - result) <= 1e-5

# Energy_throughput:
@pytest.mark.parametrize('storage_power, result',
                          [
                                ([0, -3600 * 10], 0.01),
                                ([0, 3600 * 10], 0.01),
                                ([3600 * 10, 3600 * 10], 0.02),
                                ([0, 0], 0)
                          ]
                         )
def test_energy_throughput(storage_power, result):

    system_dict = {SystemState.TIME: time_test, SystemState.AC_POWER_DELIVERED: storage_power,
                   SystemState.SYSTEM_AC_ID: 1, SystemState.SYSTEM_DC_ID: 1}
    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))

    uut = TechnicalEvaluation(system_data, analysis_config)
    assert abs(uut.energy_throughput - result) <= 1e-5

# Mean_SOC:
@pytest.mark.parametrize('soc, result',
                          [
                                ([0, 0], [0]),
                                ([0.1, 0.2], [15]),
                                ([1, 0], [50]),
                          ]
                         )
def test_mean_soc(soc, result):

    system_dict = {SystemState.TIME: time_test, SystemState.SOC: soc,
                   SystemState.SYSTEM_AC_ID: 1, SystemState.SYSTEM_DC_ID: 1}
    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))

    uut = TechnicalEvaluation(system_data, analysis_config)
    assert abs(uut.mean_soc - result) <= 1e-5

# Min_SOC:
@pytest.mark.parametrize('soc, result',
                          [
                                ([0, 0], [0]),
                                ([0.1, 0.2], [10]),
                                ([1, 0], [0]),
                          ]
                         )
def test_min_soc(soc, result):

    system_dict = {SystemState.TIME: time_test, SystemState.SOC: soc,
                   SystemState.SYSTEM_AC_ID: 1, SystemState.SYSTEM_DC_ID: 1}
    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))

    uut = TechnicalEvaluation(system_data, analysis_config)
    assert abs(uut.min_soc - result) <= 1e-5

# Max_SOC:
@pytest.mark.parametrize('soc, result',
                          [
                                ([0, 0], [0]),
                                ([0.1, 0.2], [20]),
                                ([1, 0], [100]),
                          ]
                         )
def test_max_soc(soc, result):

    system_dict = {SystemState.TIME: time_test, SystemState.SOC: soc,
                   SystemState.SYSTEM_AC_ID: 1, SystemState.SYSTEM_DC_ID: 1}
    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))

    uut = TechnicalEvaluation(system_data, analysis_config)
    assert abs(uut.max_soc - result) <= 1e-5


# FEC:
@pytest.mark.parametrize('storage_power, capacity, result',
                          [
                                ([0, 0], [10, 10], 0),
                                ([0, 3600 * 10], [10, 10], 1),
                                ([0, -3600 * 10], [10, 10], 0), # discharging is not considered
                          ]
                         )
def test_equivalent_full_cycles(storage_power, capacity, result):

    system_dict = {SystemState.TIME: time_test, SystemState.AC_POWER_DELIVERED: storage_power,
                   SystemState.CAPACITY: capacity,
                   SystemState.SYSTEM_AC_ID: 1, SystemState.SYSTEM_DC_ID: 1}
    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))

    uut = TechnicalEvaluation(system_data, analysis_config)
    assert abs(uut.equivalent_full_cycles - result) <= 1e-5


# Depth of Discharges:
@pytest.mark.parametrize('soc, result',
                          [
                                ([0.8, 0.3, 0.3, 0.3, 0.3, 0.3], 50),
                                ([0.2, 0.2, 0.3, 0.3, 0.3, 0.3], 0), # no discharge
                                ([0.6, 0.4, 0.3, 0.3, 0.3, 0.3], 30),
                                ([0.6, 0.4, 0.6, 0.4, 0.6, 0.4], 20),
                                ([0.6, 0.4, 0.6, 0.6, 0.6, 0.2], 30),
                                ([0.6, 0.0, 1.0, 0.4, 0.8, 0.2], 60),
                                ([1.0, 0.99, 0.98, 1.0, 0.99, 1.0], 1.5),
                          ]
                         )
def test_depth_of_discharges(soc, result):
    n_temp = 6
    time_test_temp = pandas.array([dstart + timestep * i for i in range(n_temp)])
    system_dict = {SystemState.TIME: time_test_temp, SystemState.SOC: soc,
                   SystemState.SYSTEM_AC_ID: 1, SystemState.SYSTEM_DC_ID: 1}
    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))

    uut = TechnicalEvaluation(system_data, analysis_config)
    assert abs(uut.depth_of_discharges - result) <= 1e-5


# Changes_of_sign:
@pytest.mark.parametrize('storage_power, result',
                          [
                                ([0, 0, 0, 0], 0),
                                ([0, -1, 0, 0], 0), # 0 is not a change of sign
                                ([0, 0, 1, 0], 0), # 0 is not a change of sign
                                ([0, 1, -1, 0], 1/3), # if 3,00003 days -> method rounds to three days
                                ([-1, 1, -1, 1], 1),
                                ([1, 1, 1, -1], 1/3)
                          ]
                         )
def test_changes_of_sign(storage_power, result):
    timestep_temp = 60*60*24+1
    n_temp = 4
    time_test_temp = pandas.array([dstart + timestep_temp * i for i in range(n_temp)])

    system_dict = {SystemState.TIME: time_test_temp, SystemState.AC_POWER_DELIVERED: storage_power,
                   SystemState.SYSTEM_AC_ID: 1, SystemState.SYSTEM_DC_ID: 1}
    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))

    uut = TechnicalEvaluation(system_data, analysis_config)
    assert abs(uut.changes_of_sign - result) <= 1e-5

# Resting times:
@pytest.mark.parametrize('storage_power, result',
                          [
                                ([0, 0, 0, 0, 0, 0], 6/60), # resting times in minutes
                                ([0, 0, 1, 0, 0, 0], 2.5/60),
                                ([0, 0, 1, 1, 0, 0], 2/60),
                                ([1, -1, -1, 1, 1, 1], 'Never in resting mode'),
                                ([-1, 1, -1, 1, 1, 0], 1/60),
                                ([0, 1, 0, -1, 0, 1], 1/60)
                          ]
                         )
def test_resting_times(storage_power, result):
    n_temp = 6
    time_test_temp = pandas.array([dstart + timestep * i for i in range(n_temp)])
    system_dict = {SystemState.TIME: time_test_temp, SystemState.AC_POWER_DELIVERED: storage_power,
                   SystemState.SYSTEM_AC_ID: 1, SystemState.SYSTEM_DC_ID: 1}
    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))

    uut = TechnicalEvaluation(system_data, analysis_config)
    if result is float:
        assert abs(uut.resting_times - result) <= 1e-5
    elif result is str:
        assert type(uut.resting_times) == str


# Energy_swapsign:
@pytest.mark.parametrize('storage_power, capacity, result',
                          [
                                ([0, 0, 0, 0, 0, 0], [10, 10, 10, 10, 10, 10], 'No energy was charged'),
                                ([0, 3600 * 10, 0, 0, 0, 0], [10, 10, 10, 10, 10, 10], 100),
                                ([-100, 3600 * 10, -100, -200, 3600 * 10, 0], [10, 10, 10, 10, 10, 10], 100),
                                ([-100, 3600 * 10, -100, 3600 * 10 *3, 0, 0], [10, 10, 10, 10, 10, 10], 200),
                                ([0, 3600 * 10, 3600 * 10, 0, 0, 0], [10, 10, 10, 10, 10, 10], 200),
                                ([0, 3600 * 10, 0, 3600 * 10, 0, 3600 * 10], [10, 10, 10, 10, 10, 10], 300),
                                ([3600 * 10, 3600 * 10, 0, 3600 * 10, 0, 3600 * 10], [10, 10, 10, 10, 10, 10], 400),
                                ([-100, 3600 * 10, 0, 3600 * 10, 0, -100], [10, 10, 10, 10, 10, 10], 200),
                          ]
                         )
def test_energy_swapsign(storage_power, capacity, result):
    n_temp = 6
    time_test_temp = pandas.array([dstart + timestep * i for i in range(n_temp)])
    system_dict = {SystemState.TIME: time_test_temp, SystemState.AC_POWER_DELIVERED: storage_power,
                   SystemState.CAPACITY: capacity,
                   SystemState.SYSTEM_AC_ID: 1, SystemState.SYSTEM_DC_ID: 1}
    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))

    uut = TechnicalEvaluation(system_data, analysis_config)
    if isinstance(result, (int, float, complex)) and not isinstance(result, bool):
        assert abs(uut.energy_swapsign - result) <= 1e-5
    elif result is str:
        assert type(uut.energy_swapsign) == str


# average_fulfillment.
@pytest.mark.parametrize('storage_fulfillment, result',
                          [
                              ([0, 0], 0),
                              ([1, 1], 100),
                              ([0, 1], 50)
                          ]
                         )
def test_average_fulfillment(storage_fulfillment, result):

    system_dict = {SystemState.TIME: time_test, SystemState.AC_FULFILLMENT: storage_fulfillment,
                   SystemState.SYSTEM_AC_ID: 1, SystemState.SYSTEM_DC_ID: 1}
    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))

    uut = TechnicalEvaluation(system_data, analysis_config)
    assert abs(uut.average_fulfillment - result) <= 1e-5

# Power electronics charge/discharge efficiencies
def setup_system_data_dict():
    n_temp = 10
    time_test_temp = pandas.array([dstart + timestep * i for i in range(n_temp)])

    dc_power_storage    = np.arange(10.0, -10.0, -2.0)
    dc_power_additional = np.zeros(10)
    dc_power = np.array([12.5, 10.0, 7.5, 5.0, 2.5, 0.0, -1.6, -3.2, -4.8, -6.4])            # eff = 0.8
    ac_power = np.array([15.625, 12.5, 9.375, 6.25, 3.125, 0.0, -1.28, -2.56, -3.84, -5.12]) # eff = 0.8
    aux_power = np.zeros(10)
    # TODO: test with a non-fix converter efficiency

    system_dict = {
        SystemState.TIME: time_test_temp, 
        SystemState.AC_POWER_DELIVERED: ac_power,
        SystemState.DC_POWER_INTERMEDIATE_CIRCUIT: dc_power,
        SystemState.DC_POWER_STORAGE: dc_power_storage,
        SystemState.DC_POWER_ADDITIONAL: dc_power_additional,
        SystemState.AUX_LOSSES: aux_power,
        SystemState.SYSTEM_AC_ID: 1, 
        SystemState.SYSTEM_DC_ID: 1
    }
    return system_dict


def test_total_acdc_efficiency_charge():
    system_dict = setup_system_data_dict()
    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))
    system_eval = SystemTechnicalEvaluation(data=system_data, config=analysis_config, path="")
    assert system_eval.total_acdc_efficiency_charge() == 80.0

def test_total_acdc_efficiency_discharge():
    system_dict = setup_system_data_dict()
    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))
    system_eval = SystemTechnicalEvaluation(data=system_data, config=analysis_config, path="")
    assert system_eval.total_acdc_efficiency_discharge() == 80.0

def test_total_dcdc_efficiency_charge():
    system_dict = setup_system_data_dict()
    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))
    system_eval = SystemTechnicalEvaluation(data=system_data, config=analysis_config, path="")
    assert system_eval.total_dcdc_efficiency_charge() == 80.0

def test_total_dcdc_efficiency_discharge():
    system_dict = setup_system_data_dict()
    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))
    system_eval = SystemTechnicalEvaluation(data=system_data, config=analysis_config, path="")
    assert system_eval.total_dcdc_efficiency_charge() == 80.0

def test_total_acdc_efficiency_charge_nan():
    system_dict = setup_system_data_dict()
    system_dict[SystemState.AC_POWER_DELIVERED]             = np.zeros(10)  # no power flow
    system_dict[SystemState.DC_POWER_INTERMEDIATE_CIRCUIT]  = np.zeros(10)
    system_dict[SystemState.DC_POWER_STORAGE]               = np.zeros(10)
    system_dict[SystemState.DC_POWER_ADDITIONAL]            = np.zeros(10)
    
    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))
    system_eval = SystemTechnicalEvaluation(data=system_data, config=analysis_config, path="")
    assert np.isnan(system_eval.total_acdc_efficiency_charge())

def test_total_charge_efficiency_inf():
    system_dict = setup_system_data_dict()
    system_dict[SystemState.AC_POWER_DELIVERED] = np.zeros(10)  # discharging, but not charging
    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))
    system_eval = SystemTechnicalEvaluation(data=system_data, config=analysis_config, path="")
    assert system_eval.total_acdc_efficiency_charge() == np.inf
    

def test_total_charge_efficiency_zero():
    system_dict = setup_system_data_dict()
    system_dict[SystemState.DC_POWER_INTERMEDIATE_CIRCUIT]  = np.zeros(10) # charging, but not discharging
    system_dict[SystemState.DC_POWER_STORAGE]               = np.zeros(10)
    system_dict[SystemState.DC_POWER_ADDITIONAL]            = np.zeros(10)

    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))
    system_eval = SystemTechnicalEvaluation(data=system_data, config=analysis_config, path="")
    assert system_eval.total_acdc_efficiency_charge() == 0.0



from simses.commons.config.analysis.analysis_config import AnalysisConfig
from simses.commons.config.analysis.general import GeneralAnalysisConfig
from simses.commons.config.generation.generator import ConfigGenerator


class AnalysisConfigGenerator(ConfigGenerator):

    """
    The AnalysisConfigGenerator is a convenience class for generating a config for SimSES analysis. Several options
    can be activated or deactivated.
    """

    def __init__(self):
        super(AnalysisConfigGenerator, self).__init__()

    def load_default_config(self) -> None:
        """
        Loads defaults config

        Returns
        -------

        """
        path: str = AnalysisConfig.CONFIG_PATH
        path += AnalysisConfig.CONFIG_NAME
        path += AnalysisConfig.DEFAULTS
        self.load_config_from(path)

    def load_local_config(self) -> None:
        """
        Loads local config

        Returns
        -------

        """
        path: str = AnalysisConfig.CONFIG_PATH
        path += AnalysisConfig.CONFIG_NAME
        path += AnalysisConfig.LOCAL
        self.load_config_from(path)

    def do_data_export(self, export: bool = True) -> None:
        """
        Analysis results will be written to files

        Parameters
        ----------
        export :
            True: data is exported, default: True

        Returns
        -------

        """
        self._set_bool(GeneralAnalysisConfig.SECTION, GeneralAnalysisConfig.EXPORT_ANALYSIS_TO_CSV, export)

    def print_results(self, printing: bool = False) -> None:
        """
        Analysis results will be printed to console

        Parameters
        ----------
        printing :
            True: results are printed, default: False

        Returns
        -------

        """
        self._set_bool(GeneralAnalysisConfig.SECTION, GeneralAnalysisConfig.PRINT_RESULTS_TO_CONSOLE, printing)

    def do_batch_analysis(self, batch: bool = False) -> None:
        """
        Multiple simulation results can be merge for comparison of results

        Parameters
        ----------
        batch :
            True: comparison data is written, default: False

        Returns
        -------

        """
        self._set_bool(GeneralAnalysisConfig.SECTION, GeneralAnalysisConfig.EXPORT_ANALYSIS_TO_BATCH, batch)

    def merge_analysis(self, merge: bool = True) -> None:
        """
        Analysis results are combined to an HTML file showing results and plots

        Parameters
        ----------
        merge :
            True: HTML file is written, default: True

        Returns
        -------

        """
        self._set_bool(GeneralAnalysisConfig.SECTION, GeneralAnalysisConfig.MERGE_ANALYSIS, merge)

    def do_technical_analysis(self, analyze: bool = True) -> None:
        """
        Provide a technical analysis of the simulation results, e.g., efficiency and aging

        Parameters
        ----------
        analyze :
            True: Technical analysis is conducted, default: True

        Returns
        -------

        """
        self._set_bool(GeneralAnalysisConfig.SECTION, GeneralAnalysisConfig.TECHNICAL_ANALYSIS, analyze)

    def do_economical_analysis(self, analyze: bool = True) -> None:
        """
        Provide an economical analysis of the simulation results, e.g., NPV and IRR (only top level system is considered)

        Parameters
        ----------
        analyze :
            True: Economical analysis is conducted, default: True

        Returns
        -------

        """
        self._set_bool(GeneralAnalysisConfig.SECTION, GeneralAnalysisConfig.ECONOMICAL_ANALYSIS, analyze)

    def do_plotting(self, plot: bool = True) -> None:
        """
        Figures are created and shown in merged HTML file

        Parameters
        ----------
        plot :
            True: Figures are plotted, default: True

        Returns
        -------

        """
        self._set_bool(GeneralAnalysisConfig.SECTION, GeneralAnalysisConfig.PLOTTING, plot)

    def do_system_analysis(self, analyze: bool = True) -> None:
        """
        Provide a technical analysis on system level

        Parameters
        ----------
        analyze :
            True: Results are analyzed, default: True

        Returns
        -------

        """
        self._set_bool(GeneralAnalysisConfig.SECTION, GeneralAnalysisConfig.SYSTEM_ANALYSIS, analyze)

    def do_lithium_ion_analysis(self, analyze: bool = True) -> None:
        """
        Provide a technical analysis for lithium-ion batteries

        Parameters
        ----------
        analyze :
            True: Results are analyzed, default: True

        Returns
        -------

        """
        self._set_bool(GeneralAnalysisConfig.SECTION, GeneralAnalysisConfig.LITHIUM_ION_ANALYSIS, analyze)

    def do_redox_flow_analysis(self, analyze: bool = True) -> None:
        """
        Provide a technical analysis for redox flow batteries

        Parameters
        ----------
        analyze :
            True: Results are analyzed, default: True

        Returns
        -------

        """
        self._set_bool(GeneralAnalysisConfig.SECTION, GeneralAnalysisConfig.REDOX_FLOW_ANALYSIS, analyze)

    def do_hydrogen_analysis(self, analyze: bool = True) -> None:
        """
        Provide a technical analysis for hydrogen systems

        Parameters
        ----------
        analyze :
            True: Results are analyzed, default: True

        Returns
        -------

        """
        self._set_bool(GeneralAnalysisConfig.SECTION, GeneralAnalysisConfig.HYDROGEN_ANALYSIS, analyze)

    def do_site_level_analysis(self, analyze: bool = True) -> None:
        """
        Provide a technical analysis on site level, e.g., comparison of site load

        Parameters
        ----------
        analyze :
            True: Results are analyzed, default: True

        Returns
        -------

        """
        self._set_bool(GeneralAnalysisConfig.SECTION, GeneralAnalysisConfig.SITE_LEVEL_ANALYSIS, analyze)



from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.technology.hydrogen.electrolyzer.thermal.abstract_model import ThermalModel


class NoThermalModel(ThermalModel):
    """This model functions at the Storage Technology Level.
      This model treats the entire Storage Technology as 1 lump.
      Current version sets temperature of Storage Technology to 298.15 K and treats it as constant"""

    def __init__(self):
        super().__init__()
        self.__temperature = 80  # °C

    def calculate(self, time, state: ElectrolyzerState, pressure_cathode_0, pressure_anode_0) -> None:
        pass

    def get_temperature(self) -> float:
        return self.__temperature

    def get_water_flow_stack(self) -> float:
        return 0.0

    def get_power_water_heating(self) -> float:
        return 0.0

    def calculate_pump_power(self, water_flow_stack: float) -> None:
        pass

    def get_pump_power(self) -> float:
        return 0.0

    def get_convection_heat(self) -> float:
        return 0.0

    def set_temperature(self, new_temperature: float):
        self.__temperature = new_temperature

    def calculate_tube_pressure_loss(self, water_flow_stack: float) -> None:
        pass

    def get_tube_pressure_loss(self) -> float:
        return 0.0



from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.cycle_detection.cycle_detector import CycleDetector
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.cyclic.cyclic_degradation import \
    CyclicDegradationModel


class GenericCellCyclicDegradationModel(CyclicDegradationModel):

    def __init__(self, cell_type: CellType, cycle_detector: CycleDetector, battery_config: BatteryConfig):
        super().__init__(cell_type, cycle_detector)
        self.__log: Logger = Logger(type(self).__name__)
        self.__total_capacity_decrease = 0
        self.__capacity_decrease = 0
        self.__initial_capacity = self._cell.get_nominal_capacity()
        self.__end_of_life_cycles = battery_config.max_equivalent_full_cycles  # FEC till EOL (=xy%) is reached. 2000 in this case
        self.__end_of_life = battery_config.eol  # EOL criteria xy%.
        self.__log.info('Considering ' + str(self.__end_of_life_cycles) + ' cycles over lifetime')

    def calculate_degradation(self, battery_state: LithiumIonState) -> None:
        # Cyclic degradation in dummy cell is linear
        self.__capacity_decrease = self._cycle_detector.get_full_equivalent_cycle() * \
                                   (1 - self.__end_of_life) / self.__end_of_life_cycles # pu

        self.__capacity_decrease = self.__capacity_decrease * self.__initial_capacity # in Ah

    def calculate_resistance_increase(self, battery_state: LithiumIonState) -> None:
        pass # No resistance increase in dummy cell model

    def get_degradation(self) -> float:
        degradation = self.__capacity_decrease - self.__total_capacity_decrease
        # Update total capacity decrease
        self.__total_capacity_decrease = self.__capacity_decrease
        return degradation

    def get_resistance_increase(self) -> float:
        return 0 # No resistance increase in dummy cell model

    def reset(self, lithium_ion_state: LithiumIonState) -> None:
        self.__capacity_decrease = 0
        self.__total_capacity_decrease = 0

    def close(self) -> None:
        self.__log.close()



from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.cycle_detection.cycle_detector import CycleDetector
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.calendar.nca_panasonicNCR import \
    PanasonicNCACalendarDegradationModel
from simses.technology.lithium_ion.degradation.cyclic.nca_panasonicNCR import \
    PanasonicNCACyclicDegradationModel
from simses.technology.lithium_ion.degradation.degradation_model import DegradationModel


class PanasonicNCADegradationModel(DegradationModel):

    def __init__(self, cell_type:CellType, cycle_detector:CycleDetector, battery_config: BatteryConfig):
        super().__init__(cell_type, PanasonicNCACyclicDegradationModel(cell_type, cycle_detector),
                         PanasonicNCACalendarDegradationModel(cell_type), cycle_detector,
                         battery_config, initial_degradation_possible=True)



from abc import ABC, abstractmethod

from simses.commons.state.technology.lithium_ion import LithiumIonState


class EquivalentCircuitModel(ABC):

    def __init__(self):
        super().__init__()

    @abstractmethod
    def update(self, time: float, battery_state: LithiumIonState) -> None:
        """Updating current, voltage, power loss and soc of lithium_ion state"""
        pass

    @abstractmethod
    def close(self) -> None:
        """Closing all resources in lithium_ion model"""
        pass



import math

import numpy as np

from simses.analysis.data.energy_management import EnergyManagementData
from simses.analysis.data.system import SystemData
from simses.analysis.evaluation.economic.revenue_stream.abstract_revenue_stream import RevenueStream
from simses.analysis.evaluation.result import EvaluationResult, Unit, Description
from simses.commons.config.analysis.economic import EconomicAnalysisConfig


class EnergyCostReduction(RevenueStream):

    def __init__(self, energy_management_data: EnergyManagementData, system_data: SystemData,
                 economic_analysis_config: EconomicAnalysisConfig):
        super().__init__(energy_management_data, system_data, economic_analysis_config)
        self.__elec_cost_per_year_storage = []  # electricity cost each year with BESS
        self.__elec_cost_per_year_base = []  # electricity cost each year without BESS
        self.__price_consumption = economic_analysis_config.electricity_price
        self.__price_generation = economic_analysis_config.pv_feed_in_tariff
        self.__Ws_to_kWh = 1/1000 * 1/3600

    def get_cashflow(self) -> np.ndarray:

        # price consumption per kWh
        price_consumption = self.__price_consumption
        # price generation per kWh
        price_generation = self.__price_generation

        time = self._energy_management_data.time
        generation = self._energy_management_data.pv_power
        load = self._energy_management_data.load_power
        storage = self._system_data.power
        delta_ts = time[1] - time[0]

        grid_power_base = load - generation
        grid_power_with_storage = load - generation + storage
        # Determine cash flow for base scenario and scenario with storage
        cashflow_base = np.asarray([power * price_consumption if power >= 0
                         else power * price_generation
                         for power in grid_power_base]) * delta_ts * self.__Ws_to_kWh
        cashflow_storage = np.asarray([power * price_consumption if power >= 0
                         else power * price_generation
                         for power in grid_power_with_storage]) * delta_ts * self.__Ws_to_kWh

        # divide into list for each project year
        self.__elec_cost_per_year_storage = self.cash_time_series_to_project_years(cashflow_storage, time)
        self.__elec_cost_per_year_base = self.cash_time_series_to_project_years(cashflow_base, time)
        revenue_nondiscounted = np.array(self.__elec_cost_per_year_base) - np.array(self.__elec_cost_per_year_storage)
        return revenue_nondiscounted

    def get_evaluation_results(self):
        key_results: [EvaluationResult] = list()
        key_results.append(EvaluationResult(Description.Economical.SCI.COST_WITHOUT_STORAGE, Unit.EURO, self.__elec_cost_per_year_base))
        key_results.append(EvaluationResult(Description.Economical.SCI.COST_WITH_STORAGE, Unit.EURO, self.__elec_cost_per_year_storage))
        return key_results

    def get_assumptions(self):
        assumptions: [EvaluationResult] = list()
        assumptions.append(EvaluationResult(Description.Economical.SCI.COST_ELECTRICITY, Unit.EURO_PER_KWH, self.__price_consumption))
        assumptions.append(EvaluationResult(Description.Economical.SCI.PV_FEED_IN_TARIFF, Unit.EURO_PER_KWH, self.__price_generation))
        return assumptions

    def close(self):
        pass




from abc import ABC, abstractmethod

from simses.commons.timeseries.timevalue import TimeValue


class Interpolation(ABC):
    """
    Interpolation implementations are to interpolate values between values as TimeValue objects. Each implementation
    has a specific interpolation behaviour (linear, splines, ...).
    """

    def __init__(self):
        super().__init__()

    @staticmethod
    def is_necessary(tstmp: float, data: [TimeValue]) -> bool:
        if len(data) == 2:
            return data[-2].time < tstmp <= data[-1].time
        return False

    @abstractmethod
    def interpolate(selfself, time: float, recent: TimeValue, last: TimeValue) -> float:
        """
        Interpolate values of given TimeValue objects to time

        Parameters
        ----------
        time :
            timestamp in s which lies in between recent and last
        recent :
            TimeValue with a timestamp >= time
        last :
            TimeValue with a timestamp <= time

        Returns
        -------
        float:
            interpolated value
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """
        Closing resources in interpolation

        Returns
        -------

        """
        pass



import pandas

from simses.analysis.data.abstract_data import Data
from simses.analysis.utils import get_sum_for, get_positive_values_from
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.constants import Hydrogen
from simses.commons.state.technology.hydrogen import HydrogenState


class HydrogenData(Data):
    """
    DEPRECATED - Provides time series data from HydrogenState
    """

    def __init__(self, config: GeneralSimulationConfig, data: pandas.DataFrame):
        super().__init__(config, data)
        self.__config = config

    @property
    def id(self) -> str:
        return str(int(self._get_first_value(HydrogenState.SYSTEM_AC_ID))) + '.' + \
               str(int(self._get_first_value(HydrogenState.SYSTEM_DC_ID)))

    @property
    def time(self):
        return self._get_data(HydrogenState.TIME)

    @property
    def power(self):
        return self._get_data(HydrogenState.POWER)

    @property
    def dc_power(self):
        return self.power

    @property
    def energy_difference(self):
        soc = self._get_difference(HydrogenState.SOC)
        capacity = self.initial_capacity
        return soc * capacity

    @property
    def soc(self):
        return self._get_data(HydrogenState.SOC)

    @property
    def capacity(self):
        return self._get_data(HydrogenState.CAPACITY) / 1000.0

    @property
    def state_of_health(self):
        return self._get_data(HydrogenState.SOH)

    @property
    def fulfillment(self):
        return self._get_data(HydrogenState.FULFILLMENT)

    @property
    def current_el(self):
        return self._get_data(HydrogenState.CURRENT_EL)

    @property
    def current_density_el(self):
        return self._get_data(HydrogenState.CURRENT_DENSITY_EL)

    @property
    def ref_voltage_el(self):
        return self._get_data(HydrogenState.REFERENCE_VOLTAGE_EL)

    @property
    def hydrogen_outflow(self):
        return self._get_data(HydrogenState.HYDROGEN_OUTFLOW)

    @property
    def hydrogen_production(self):
        return self._get_data(HydrogenState.HYDROGEN_PRODUCTION)

    @property
    def pressure_anode_el(self):
        return self._get_data(HydrogenState.PRESSURE_ANODE_EL)

    @property
    def pressure_cathode_el(self):
        return self._get_data(HydrogenState.PRESSURE_CATHODE_EL)

    @property
    def part_pressure_h2_el(self):
        return self._get_data(HydrogenState.PART_PRESSURE_H2_EL)

    @property
    def part_pressure_o2_el(self):
        return self._get_data(HydrogenState.PART_PRESSURE_O2_EL)

    @property
    def sat_pressure_h20_el(self):
        return  self._get_data(HydrogenState.SAT_PRESSURE_H2O_EL)

    @property
    def tank_pressure(self):
        return self._get_data(HydrogenState.TANK_PRESSURE)

    @property
    def temperature_el(self):
        return self._get_data(HydrogenState.TEMPERATURE_EL)

    @property
    def storage_fulfillment(self):
        return self._get_data(HydrogenState.FULFILLMENT)

    @property
    def convection_heat(self):
        return self._get_data(HydrogenState.CONVECTION_HEAT)

    @property
    def power_water_heating_el(self):
        return self._get_data(HydrogenState.POWER_WATER_HEATING_EL)

    @property
    def power_pump_el(self):
        return self._get_data(HydrogenState.POWER_PUMP_EL)

    @property
    def power_gas_drying(self):
        return self._get_data(HydrogenState.POWER_GAS_DRYING)

    @property
    def power_compressor(self):
        return self._get_data(HydrogenState.POWER_COMPRESSOR)

    @property
    def total_h2_production(self):
        return self._get_data(HydrogenState.TOTAL_HYDROGEN_PRODUCTION)

    @property
    def resistance_increase_cyclic_el(self):
        return self._get_data(HydrogenState.RESISTANCE_INCREASE_CYCLIC_EL)

    @property
    def resistance_increase_calendar_el(self):
        return self._get_data(HydrogenState.RESISTANCE_INCREASE_CALENDAR_EL)

    @property
    def resistance_increase_el(self):
        return self._get_data(HydrogenState.RESISTANCE_INCREASE_EL)

    @property
    def exchange_current_dens_decrease_cyclic_el(self):
        return self._get_data(HydrogenState.EXCHANGE_CURRENT_DENS_DECREASE_CYCLIC_EL)

    @property
    def exchange_current_dens_decrease_calendar_el(self):
        return self._get_data(HydrogenState.EXCHANGE_CURRENT_DENS_DECREASE_CALENDAR_EL)

    @property
    def exchange_current_dens_decrease_el(self):
        return self._get_data(HydrogenState.EXCHANGE_CURRENT_DENS_DECREASE_EL)

    @property
    def soh_el(self):
        return self._get_data(HydrogenState.SOH_EL)

    @property
    def total_energy_pump_el(self):
        return get_sum_for(self.power_pump_el) * self.convert_watt_to_kWh

    @property
    def total_energy_compressor(self):
        return get_sum_for(self.power_compressor) * self.convert_watt_to_kWh

    @property
    def total_energy_gas_drying(self):
        return get_sum_for(self.power_gas_drying) * self.convert_watt_to_kWh

    @property
    def total_energy_heating(self):
        return get_sum_for(get_positive_values_from(self.power_water_heating_el)) * self.convert_watt_to_kWh

    @property
    def total_energy_reaction(self):
        return get_sum_for(self.power) * self.convert_watt_to_kWh

    @property
    def total_energy_h2_lhv(self):
        return get_sum_for(
            self.hydrogen_outflow) * self.__config.timestep * 2 * Hydrogen.MOLAR_MASS_HYDROGEN * Hydrogen.LOWER_HEATING_VALUE

    @property
    def total_amount_h2_kg(self):
        return get_sum_for(self.hydrogen_outflow) * self.__config.timestep * 2 * Hydrogen.MOLAR_MASS_HYDROGEN

    @property
    def total_amount_h2_nqm(self):
        return get_sum_for(self.hydrogen_outflow) * self.__config.timestep * 2 * \
               Hydrogen.MOLAR_MASS_HYDROGEN * Hydrogen.NORM_QUBIC_M

    @classmethod
    def get_system_data(cls, path: str, config: GeneralSimulationConfig) -> list:
        system_data: [pandas.DataFrame] = cls._get_system_data_for(path, HydrogenState, HydrogenState.TIME,
                                                                   HydrogenState.SYSTEM_AC_ID,
                                                                   HydrogenState.SYSTEM_DC_ID)
        res: [HydrogenData] = list()
        for data in system_data:
            res.append(HydrogenData(config, data))
        return res





from abc import ABC, abstractmethod


class DataHandler(ABC):
    """
    Abstract Base Class for Data Export
    """

    def __init__(self):
        pass

    @abstractmethod
    def start(self) -> None:
        """
        Function to write data from queue to csv.

        Parameters
        ----------

        Returns
        -------

        """
        pass

    @abstractmethod
    def transfer_data(self, data: dict) -> None:
        """
        Function to place data into the queue. This data is picked up by the run function and written into a CSV file_name.

        Parameters
        ----------
        data : list

        Returns
        -------

        """
        pass

    @abstractmethod
    def simulation_done(self) -> None:
        """
        Function to let the DataExport Thread know that the simulation is done and stop after emptying the queue.

        Parameters
        ----------

        Returns
        -------

        """
        pass

    @abstractmethod
    def close(self) -> None:
        """
        Closing all open resources in data export

        Parameters
        ----------

        Returns
        -------

        """
        pass

    def is_alive(self) -> None:
        """
        Checks if a thread is running

        Parameters
        ----------

        Returns
        -------

        """
        pass


from configparser import ConfigParser
from simses.commons.data.data_handler import DataHandler
from simses.commons.error import EndOfLifeError
from simses.commons.state.abstract_state import State
from simses.commons.state.parameters import SystemParameters
from simses.commons.state.system import SystemState
from simses.logic.power_distribution.power_distributor import PowerDistributor
from simses.system.factory import StorageSystemFactory
from simses.system.storage_system_ac import StorageSystemAC


class StorageCircuit:

    """
    StorageCircuit is the top level class including all AC storage systems. The is distributed via a PowerDistributor
    logic to each AC storage system.
    """

    def __init__(self, data_export: DataHandler, config: ConfigParser):
        factory: StorageSystemFactory = StorageSystemFactory(config)
        self.__storage_systems: [StorageSystemAC] = factory.create_storage_systems_ac(data_export)
        self.__power_distributor: PowerDistributor = factory.create_power_distributor_ac()
        factory.close()

    def update(self, time: float, power: float, power_dist: [float] = None) -> None:
        if power_dist is None:
            power_dist = []
        states: [State] = list()
        for system in self.__storage_systems:  # type: StorageSystemAC
            states.append(system.state)
        self.__power_distributor.set(time, states, power)
        end_of_life_reached: bool = False
        for i in range(0, len(self.__storage_systems)):
            if not power_dist:
                local_power: float = self.__power_distributor.get_power_for(power, self.__storage_systems[i].state)
            else:
                local_power: float = power_dist[i]
            try:
                self.__storage_systems[i].update(local_power, time)
            except EndOfLifeError:
                end_of_life_reached = True
        if end_of_life_reached:
            raise EndOfLifeError()

    @property
    def state(self) -> SystemState:
        system_states = list()
        for storage in self.__storage_systems:  # type: StorageSystemAC
            system_states.append(storage.state)
        system_state: SystemState = SystemState.sum_parallel(system_states)
        system_state.set(SystemState.SYSTEM_AC_ID, 0)
        system_state.set(SystemState.SYSTEM_DC_ID, 0)
        return system_state

    @property
    def ac_system_states(self) -> [SystemState]:
        system_states = list()
        for storage in self.__storage_systems:  # type: StorageSystemAC
            system_states.append(storage.state)
        return system_states

    def get_system_parameters(self) -> dict:
        parameters: dict = dict()
        subsystems: list = list()
        for system in self.__storage_systems:  # type: StorageSystemAC
            subsystems.append(system.get_system_parameters())
        parameters[SystemParameters.POWER_DISTRIBUTION] = type(self.__power_distributor).__name__
        parameters[SystemParameters.SUBSYSTEM] = subsystems
        return {SystemParameters.PARAMETERS: parameters}

    def reset_profiles(self, ts_adapted: float) -> None:
        """
        Enables looping of the simulation beyond the original length of the time series for the AmbientThermalModel and
        SolarIrradiationModel
        """
        for system in self.__storage_systems:  # type: StorageSystemAC
            system.reset_profiles(ts_adapted)

    def close(self) -> None:
        """Closing all resources in storage systems"""
        self.__power_distributor.close()
        for storage in self.__storage_systems:  # type: StorageSystemAC
            storage.close()



from simses.commons.profile.file import FileProfile
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.system.auxiliary.auxiliary import Auxiliary
from simses.system.dc_coupling.load.dc_load import DcLoad


class DcBusChargingProfile(DcLoad):

    """
    DcBusChargingProfile is able to read in SOC profiles and calculates a load power for charging.
    """

    def __init__(self, config: GeneralSimulationConfig, capacity: float, file_name: str):
        super().__init__()
        self.__profile: FileProfile = FileProfile(config, file_name, value_index=2)
        self.__power: float = 0.0
        self.__capacity: float = capacity
        self.__last_soc: float = self.__profile.next(config.start)
        self.__timestep: float = config.timestep
        self.__Wh_to_Ws: float = 3600.0
        self.__charge_efficiency: float = 0.98

    def get_power(self) -> float:
        return self.__power

    def calculate_power(self, time: float) -> None:
        soc = self.__profile.next(time)
        power = (soc - self.__last_soc) * self.__capacity * self.__Wh_to_Ws / self.__timestep / self.__charge_efficiency
        self.__power = max(power, 0.0)
        self.__last_soc = soc

    def get_auxiliaries(self) -> [Auxiliary]:
        return list()

    def close(self) -> None:
        pass



from configparser import ConfigParser

from simses.commons.data.data_handler import DataHandler
from simses.commons.state.technology.fuel_cell import FuelCellState
from simses.system.auxiliary.pump.abstract_pump import Pump
from simses.system.auxiliary.pump.fixeta_centrifugal import FixEtaCentrifugalPump
from simses.system.auxiliary.water_heating.water_heater import WaterHeating
from simses.technology.hydrogen.fuel_cell.factory import FuelCellFactory
from simses.technology.hydrogen.fuel_cell.pressure.pressure_model import PressureModel
from simses.technology.hydrogen.fuel_cell.stack.no_fuel_cell import NoFuelCell
from simses.technology.hydrogen.fuel_cell.stack.stack_model import \
    FuelCellStackModel
from simses.technology.hydrogen.fuel_cell.thermal.thermal_model import ThermalModel


class FuelCell:

    """
    FuelCell is the top level class incorporating a FuelCellStackModel, a ThermalModel and a PressureModel. The
    specific classes are instantiated within the FuelCellFactory.
    """

    __PUMP_EFFICIENCY: float = 0.7

    def __init__(self, system_id: int, storage_id: int, fuel_cell: str, max_power: float,
                 temperature: float, config: ConfigParser, data_handler: DataHandler):
        super().__init__()
        self.__data_handler: DataHandler = data_handler
        self.__pump: Pump = FixEtaCentrifugalPump(self.__PUMP_EFFICIENCY)
        self.__water_heating: WaterHeating = WaterHeating()
        factory: FuelCellFactory = FuelCellFactory(config)
        self.__stack_model: FuelCellStackModel = factory.create_fuel_cell_stack(fuel_cell, max_power)
        self.__pressure_model: PressureModel = factory.create_pressure_model(self.__stack_model)
        self.__thermal_model: ThermalModel = factory.create_thermal_model(self.__stack_model, temperature)
        self.__state: FuelCellState = factory.create_state(system_id, storage_id, temperature, self.__stack_model)
        factory.close()
        self.__log_data: bool = not isinstance(self.__stack_model, NoFuelCell)
        self.__write_data()

    def update(self, time: float, power: float) -> None:
        """
        Updates current, voltage and hydrogen flow of hydrogen state

        Parameters
        ----------
        time :
        state :
        power :

        Returns
        -------

        """
        local_power = power if power < 0.0 else 0.0
        state: FuelCellState = self.__state
        self.__stack_model.update(local_power, state)
        self.__pressure_model.update(time, state)
        self.__thermal_model.update(time, state)
        state.time = time
        self.__write_data()

    def __write_data(self) -> None:
        if self.__log_data:
            self.__data_handler.transfer_data(self.__state.to_export())

    def get_auxiliaries(self) -> list:
        return list()

    @property
    def state(self) -> FuelCellState:
        return self.__state

    def close(self):
        self.__stack_model.close()
        self.__pressure_model.close()
        self.__thermal_model.close()
        self.__pump.close()
        self.__water_heating.close()



import math

from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.calendar.calendar_degradation import \
    CalendarDegradationModel


class GenericCellCalendarDegradationModel(CalendarDegradationModel):

    def __init__(self, cell_type: CellType, battery_config: BatteryConfig):
        super().__init__(cell_type)
        self.__log: Logger = Logger(type(self).__name__)
        self.__total_capacity_decrease = 0
        self.__capacity_decrease = 0
        self.__total_time_passed = 0
        self.__initial_capacity = self._cell.get_nominal_capacity()
        self.__end_of_life_duration = battery_config.max_calendar_lifetime * 365. * 24. * 60. * 60.
        self.__end_of_life = battery_config.eol # EOL criteria xy%.
        self.__log.info('Considering ' + str(self.__end_of_life_duration) + ' years over lifetime')

    def calculate_degradation(self, time: float, battery_state: LithiumIonState) -> None:
        # Calendar degradation in dummy cell is a function of sqrt(t)
        self.__total_time_passed += time - battery_state.time
        self.__capacity_decrease = math.sqrt(self.__total_time_passed) * (1 - self.__end_of_life) / \
                                   math.sqrt(self.__end_of_life_duration) # in p.u

        self.__capacity_decrease = self.__capacity_decrease * self.__initial_capacity

    def calculate_resistance_increase(self, time: float, battery_state: LithiumIonState) -> None:
        pass # No resistance increase in dummy cell model

    def get_degradation(self) -> float:
        degradation = self.__capacity_decrease - self.__total_capacity_decrease
        # Update total capacity decrease
        self.__total_capacity_decrease = self.__capacity_decrease
        return degradation

    def get_resistance_increase(self) -> float:
        return 0 # No resistance increase in dummy cell model

    def reset(self, battery_state: LithiumIonState) -> None:
        self.__capacity_decrease = 0
        self.__total_capacity_decrease = 0
        self.__total_time_passed = 0

    def close(self) -> None:
        self.__log.close()



import os
import sys
import time
from configparser import ConfigParser
from math import floor
from multiprocessing import Queue
from queue import Full

from simses.commons.config.abstract_config import Config
from simses.commons.config.data.data_config import DataConfig
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.console_printer import ConsolePrinter
from simses.commons.data.csv_data_handler import CSVDataHandler
from simses.commons.data.no_data_handler import NoDataHandler
from simses.commons.error import EndOfLifeError
from simses.commons.log import Logger
from simses.commons.state.parameters import SystemParameters
from simses.commons.state.system import SystemState
from simses.commons.utils.utilities import format_float
from simses.logic.energy_management.energy_management_system import EnergyManagement
from simses.system.storage_circuit import StorageCircuit


class StorageSimulation:

    """
    StorageSimulation constructs the the storage systems and energy management system in order to execute the simulation.
    In the run() method the timestamp for the simulation is advanced as configured. Alternatively, simulation is included
    in another framework advancing timestamps itself, e.g. run_one_step() or evaluate_multiple_steps(). StorageSimulation
    also provided information of the current status of the simulation to the user.
    """

    def __init__(self, path: str, config: ConfigParser, printer_queue: Queue):
        """
        Constructor of StorageSimulation

        Parameters
        ----------
        path :
            path to result folder
        config :
            Optional configs taken into account overwriting values from provided config file
        printer_queue :
            Optional queue for concurrent simulation process for providing progress status of simulations
        """
        self.__path = path
        self.__log = Logger(type(self).__name__)
        # Only instantiated in order write data config to results folder, maybe this way should be improved
        self.__data_config: Config = DataConfig(None, None)
        self.__config = GeneralSimulationConfig(config)
        if self.__config.export_data:
            self.__data_export = CSVDataHandler(path, self.__config)
        else:
            self.__data_export = NoDataHandler()
        self.__energy_management: EnergyManagement = EnergyManagement(self.__data_export, config)
        self.__storage_system = StorageCircuit(self.__data_export, config)
        self.__name: str = os.path.basename(os.path.dirname(self.__path))
        self.__printer_queue: Queue = printer_queue
        self.__send_register_signal()
        self.__max_loop = self.__config.loop
        self.__start = self.__config.start
        self.__end = self.__config.end
        self.__timestep = self.__config.timestep  # sec
        # duration to the last executed time step
        self.__duration = floor((self.__end - self.__start) / self.__timestep) * self.__timestep
        self.__timesteps_per_hour = 3600 / self.__timestep
        system_parameters: SystemParameters = SystemParameters()
        system_parameters.set_all(self.__storage_system.get_system_parameters())
        system_parameters.write_parameters_to(path)

    def run(self) -> None:
        """
        Executes simulation

        Returns
        -------

        """
        self.__log.info('start')
        sim_start = time.time()
        ts_performance = []
        try:
            loop = 0
            ts = self.__start
            while loop < self.__max_loop:
                self.__log.info('Loop: ' + str(loop))
                ts_adapted = loop * self.__duration
                while ts <= (ts_adapted + self.__end) - self.__timestep:
                    ts_before = time.time()
                    ts += self.__timestep
                    self.run_one_step(ts, ts_adapted)
                    ts_performance.append(time.time() - ts_before)
                    self.__print_progress(ts)
                loop += 1
                if loop < self.__max_loop:
                    ts_adapted = loop * self.__duration
                    self.reset_profiles(ts_adapted)
        except EndOfLifeError as err:
            self.__log.error(err)
        finally:
            self.close()
            self.__print_end(ts_performance, sim_start)

    def __print_progress(self, tstmp: float) -> None:
        progress = (tstmp - self.__start) / (self.__duration * self.__max_loop) * 100
        line: str = '|%-20s| ' % ('#' * round(progress / 5)) + format_float(progress, 1) + '%'
        output: dict = {self.__name: line}
        if self.__printer_queue is None:
            sys.stdout.write('\r' + str(output))
            sys.stdout.flush()
        self.__put_to_queue(output)

    def __put_to_queue(self, output: dict, blocking: bool = False) -> None:
        if self.__printer_queue is not None:
            try:
                if blocking:
                    self.__printer_queue.put(output)
                else:
                    self.__printer_queue.put_nowait(output)
            except Full:
                return

    def __send_stop_signal(self) -> None:
        self.__put_to_queue({self.__name: ConsolePrinter.STOP_SIGNAL}, blocking=True)

    def __send_register_signal(self) -> None:
        self.__put_to_queue({self.__name: ConsolePrinter.REGISTER_SIGNAL}, blocking=True)

    def __print_end(self, ts_performance: list, sim_start: float) -> None:
        try:
            sim_end = time.time()
            duration: str = format_float(sim_end - sim_start)
            duration_per_step: str = format_float(sum(ts_performance) * 1000 / len(ts_performance))
            self.__log.info('100.0% done. Duration in sec: ' + duration)
            self.__log.info('Duration per step in ms:      ' + duration_per_step)
            if self.__printer_queue is None:
                print('\r[' + self.__name + ': |%-20s| ' % ('#' * 20) + '100.0%]')
                print('          Duration in s: ' + duration)
                print('Duration per step in ms: ' + duration_per_step)
        except ZeroDivisionError:
            self.__log.warn('No performance indicators could be calculated.')

    def run_one_step(self, ts: float, ts_adapted: float = 0, power: float = None, power_dist: [float] = None) -> None:
        """
        Advances simulation for one step. Results can be obtained via state property.

        Parameters
        ----------
        ts :
            next timestamp in s
        ts_adapted :
            timestamp adaption for looping simulations multiple times (should only be used with stand alone SimSES)
        power :
            next power transfered to storage system in W, if None power is taken from configured energy management
        power_dist :
            next power distribution in W for every AC system as a list. If None, power is taken from power distributor classes.
        Returns
        -------

        """
        if power_dist is None:
            power_dist = []
        state = self.__storage_system.state
        if not self.__data_export.is_alive():
            self.__data_export.start()
        self.__data_export.transfer_data(state.to_export())
        if power is None:
            power = self.__energy_management.next(ts - ts_adapted, state)
        try:
            self.__storage_system.update(ts, power, power_dist)
        finally:
            self.__energy_management.export(ts)

    def evaluate_multiple_steps(self, start: float, timestep: float, power: list) -> [SystemState]:
        """
        Runs multiple steps of the simulation with the given start time, timestep and power list.
        If no power list is provided, the simulation will not be advanced.

        Parameters
        ----------
        start :
            start time in s
        timestep :
            timestep in s
        power :
            list of power for each timestep in W

        Returns
        ----------
        list:
            Returns a list of system states for each timestep

        """
        res: [SystemState] = list()
        ts = start
        for pow in power:
            self.run_one_step(ts=ts, power=pow)
            res.append(self.state)
            ts += timestep
        return res

    @property
    def state(self) -> SystemState:
        """

        Returns
        -------
        SystemState:
            current state of top level system
        """
        return self.__storage_system.state

    @property
    def ac_system_states(self) -> [SystemState]:
        """

        Returns
        -------
        SystemState:
            list of current states for all AC systems
        """
        return self.__storage_system.ac_system_states

    def reset_profiles(self, ts_adapted: float) -> None:
        """
        Enables looping of the simulation beyond the original length of the time series for all profile-based modules
        such as the EnergyManagement, the AmbientThermalModel, and SolarIrradiationModel
        """
        self.__energy_management.close()
        self.__energy_management: EnergyManagement = self.__energy_management.create_instance()
        self.__storage_system.reset_profiles(ts_adapted)

    def close(self) -> None:
        """
        Closing all resources of simulation

        Returns
        -------

        """
        self.__log.info('closing')
        self.__data_export.transfer_data(self.__storage_system.state.to_export())
        self.__send_stop_signal()
        self.__config.write_config_to(self.__path)
        self.__data_config.write_config_to(self.__path)
        self.__log.close()
        self.__data_export.close()
        self.__energy_management.close()
        self.__storage_system.close()



import logging
import os

from simses.commons.config.log import LogConfig


class Logger:

    """
    SimSES Logger with possibility to write logs to console and file. Logger is configurable via logger.ini in config.
    """

    __format: str = '%(asctime)s %(levelname)s [%(name)s] %(message)s'
    __filename: str = 'simses.log'
    __path: str = os.getcwd() + '/'

    def __init__(self, name: str, log_level: int = 0):
        log_config: LogConfig = LogConfig()
        log_level = log_config.log_level if log_level == 0 else log_level
        # logger and handler
        logger = logging.getLogger(name.split('.')[-1].upper().strip('_'))
        file_handler = logging.FileHandler(self.__path + self.__filename)
        stream_handler = logging.StreamHandler()
        # setting log level
        logger.setLevel(log_level)
        file_handler.setLevel(log_level)
        stream_handler.setLevel(log_level)
        # setting format
        formatter = logging.Formatter(fmt=self.__format)
        file_handler.setFormatter(formatter)
        stream_handler.setFormatter(formatter)
        # add handler
        if log_config.log_to_console:
            logger.addHandler(stream_handler)
        if log_config.log_to_file:
            logger.addHandler(file_handler)
        self.logger: logging.Logger = logger

    def info(self, msg) -> None:
        self.logger.info(msg)

    def warn(self, msg) -> None:
        self.logger.warning(msg)

    def error(self, msg) -> None:
        self.logger.error(msg)

    def debug(self, msg) -> None:
        self.logger.debug(msg)

    def close(self) -> None:
        handlers = self.logger.handlers[:]
        for handler in handlers:
            handler.close()
            self.logger.removeHandler(handler)



from abc import abstractmethod, ABC

from simses.commons.state.technology.fuel_cell import FuelCellState


class PressureModel(ABC):
    """
    PressureModelEl calculates the pressure development at the cathode of the electrolyzer in dependency of the
    hydrogenproductionrate, the hydrogenoutflow and the temperature
    """

    # MOLAR_MASS_HYDROGEN = 1.00794 * 10 ** (-3) # kg/mol
    # FARADAY_CONST: float = 96485.3321  # As/mol
    # IDEAL_GAS_CONST: float = 8.314462  # J/(mol K)
    # HEAT_CAPACITY_WATER: float = 4184  # J/(kg K)
    # MOLAR_MASS_WATER: float = 0.018015  # kg/mol
    # HYDROGEN_ISENTROP_EXPONENT = 1.0498  # # from: "Wasserstoff in der Fahrzeugtechnik"
    # HYDROGEN_REAL_GAS_FACTOR = 1  # valid for pressures << 13 bar and temperatures >> 33 K

    def __init__(self):
        super().__init__()

    def update(self, time: float, state: FuelCellState):
        self.calculate(time, state)
        state.pressure_cathode_fc = self.get_pressure_cathode_fc()  # bar
        state.pressure_anode_fc = self.get_pressure_anode_fc()  # bar   anode pressure is not varied -> stays constant
        state.hydrogen_use = self.get_h2_inflow()
        state.oxygen_use = self.get_o2_inflow()

    @abstractmethod
    def calculate(self, time, state: FuelCellState) -> None:
        pass

    @abstractmethod
    def get_pressure_cathode_fc(self) -> float:
        pass

    @abstractmethod
    def get_pressure_anode_fc(self) -> float:
        pass

    @abstractmethod
    def get_h2_inflow(self) -> float:
        pass

    @abstractmethod
    def get_o2_inflow(self) -> float:
        pass

    def close(self) -> None:
        """ Closing all resources in pressure model """
        pass



from abc import ABC, abstractmethod

import numpy

from simses.analysis.data.energy_management import EnergyManagementData
from simses.analysis.data.system import SystemData
from simses.analysis.evaluation.result import EvaluationResult
from simses.commons.config.analysis.economic import EconomicAnalysisConfig


class RevenueStream(ABC):
    """Calculates the cashflow for each project year that is generated by the given RevenueStream.
    A RevenueStream can be seen as the financial impact of the various applications of a BESS such as peak shaving,
    self-consumption, frequency containment reserve, etc.. """

    def __init__(self, energy_management_data: EnergyManagementData, system_data: SystemData,
                 economic_analysis_config: EconomicAnalysisConfig):
        self._energy_management_data: EnergyManagementData = energy_management_data
        self._system_data: SystemData = system_data
        self._economic_analysis_config: EconomicAnalysisConfig = economic_analysis_config
        self._investment_cost: float = 0.0
        self._seconds_per_year = 24 * 60 * 60 * 365

    @abstractmethod
    def get_cashflow(self) -> numpy.ndarray:
        """
        Returns non-discounted cashflow for every project year for the given RevenueStream.

        Returns
        -------
        list of float
            List of cashflow for every project year
        """
        pass

    @abstractmethod
    def get_evaluation_results(self) -> [EvaluationResult]:
        """
        Returns list of EvaluationResults for the given RevenueStream.

        Returns
        -------
        list of EvaluationResult
            List of EvaluationResults
        """
        pass

    @abstractmethod
    def get_assumptions(self) -> [EvaluationResult]:
        """
        Returns list of assumptions for the given RevenueStream. Assumptions are handled with the
        EvaluationResults class.

        Returns
        -------
        list of EvaluationResult
            List of EvaluationResults, representing assumptions
        """
        pass

    def set_investment_cost(self, value: float) -> None:
        self._investment_cost = value

    def cash_time_series_to_project_years(self, cashflows: [float], time: [float]) -> [float]:
        """
        Converts a cashflow time series for every timestep to a cashflow time series for every project year.
        A project year is assumed to be 365 days long. A full leap year would therefore be one project year and one day.

        Parameters
        ----------
        cashflows :
            List of float, representing cashflow for each timestep
        time :
            List of float, unix timestamps for each simulation step

        Returns
        -------
        list of float
            List of float, representing cashflow for each project year.
        """
        cashflow_project_years = []
        start_time = time[0]
        cashflow_sum = 0
        add_last = True
        for (t, c) in zip(time, cashflows):
            if t < start_time + self._seconds_per_year:
                cashflow_sum += c
                add_last = True
            else:
                cashflow_project_years.append(cashflow_sum)
                cashflow_sum = 0
                cashflow_sum += c
                start_time = t
                add_last = False
        if add_last:
            cashflow_project_years.append(cashflow_sum)
        return cashflow_project_years

    @abstractmethod
    def close(self):
        pass



from simses.commons.state.technology.fuel_cell import FuelCellState
from simses.technology.hydrogen.fuel_cell.thermal.thermal_model import ThermalModel


class NoThermalModel(ThermalModel):
    """This model functions at the Storage Technology Level.
      This model treats the entire Storage Technology as 1 lump.
      Current version sets temperature of Storage Technology to 298.15 K and treats it as constant"""

    def __init__(self):
        super().__init__()
        self.__temperature = 298.15 # K

    def calculate(self, time, state: FuelCellState) -> None:
        self.__temperature = state.temperature

    def get_temperature(self) -> float:
        return self.__temperature

    def get_pump_power(self) -> float:
        return 0.0

    def get_water_flow_stack(self) -> float:
        return 0.0

    def get_power_water_heating(self) -> float:
        return 0.0



from simses.commons.config.simulation.electrolyzer import ElectrolyzerConfig
from simses.technology.hydrogen.control.pressure.pressure_controller import \
    PressureController


class IdealVarCathodePressureController(PressureController):
    """ This pressure controller controls the cathode pressure at a disired level and keeps the anode pressure
    at ambient level"""
    def __init__(self, config: ElectrolyzerConfig):
        super().__init__()
        self.__max_pressure_variation_rate: float = 0.5  # bar/s
        self.__pressure_control_on: bool = False
        self.__no_h2_production_counter: float = 0
        self.__shut_down_time: float = 0.5  # h
        self.__pressure_cathode_desire: float = config.desire_pressure_cathode
        self.__pressure_anode_desire: float = config.desire_pressure_anode

    def calculate_n_h2_out(self, pressure_cathode, n_h2_prod, timestep, pressure_factor) -> float:
        """
        calculates the necessary hydrogen outflow to reach the desire pressure relative to the actual hydrogen
        production and cathode pressure

        input: pressure cathode in barg, pressure cathode desire in barg, h2 production in mol/s, timestep in s,
        pressure factor in bar/mol

        output: h2 outflow in mol/s
        """
        delta_pressure = self.__pressure_cathode_desire - pressure_cathode  # bar
        pressure_variation_rate_desire = delta_pressure / timestep  # bar/s  >0 if pressure cathode > pressure cathode deisire
        self.__pressure_control_on = self.__check_pressure_control_on(n_h2_prod, timestep)
        if self.__pressure_control_on:  # control is on -> pressure value is kept at its desire value
            if abs(pressure_variation_rate_desire) < self.__max_pressure_variation_rate:
                pressure_variation_rate = pressure_variation_rate_desire
            else:
                pressure_variation_rate = pressure_variation_rate_desire / abs(pressure_variation_rate_desire) * self.__max_pressure_variation_rate
            # p_c_1 = pressure_cathode + pressure_variation_rate * timestep
            n_h2_out = n_h2_prod - pressure_variation_rate / pressure_factor
            return max(n_h2_out, 0)
        else:
            p_c_1 = max(pressure_cathode - self.__max_pressure_variation_rate * timestep, 0)
            return n_h2_prod - (p_c_1 - pressure_cathode) / (pressure_factor * timestep)

    def calculate_n_o2_out(self, pressure_anode: float, n_o2_prod: float, timestep: float) -> float:
        if n_o2_prod >= 0:
            return n_o2_prod
        else:
            return 0  # no intake of atmosphere in case of negative oxygen production (-> only permeation)

    def __check_pressure_control_on(self, n_h2_prod: float, timestep: float) -> bool:
        if n_h2_prod <= 0:
            self.__no_h2_production_counter += 1
        else:
            self.__no_h2_production_counter = 0
        if self.__no_h2_production_counter <= self.__shut_down_time * 3600 / timestep:
            return True
        else:
            return False

    def calculate_n_h2_in(self, pressure_anode: float, n_h2_used: float) -> float:
        return n_h2_used

    def calculate_n_o2_in(self, pressure_cathode: float, n_o2_used: float) -> float:
        return n_o2_used



import os
import shutil
import time
from configparser import ConfigParser
from simses.commons.config.abstract_config import Config
from simses.commons.config.generation.analysis import AnalysisConfigGenerator
from simses.commons.config.simulation.simulation_config import SimulationConfig
from simses.commons.utils.utilities import remove_file
from simses.simulation.batch_processing import BatchProcessing


class SystemTest(BatchProcessing):

    """
    TestBatchProcessing execute system tests for various system configurations.
    """

    __TEST_NAME: str = 'test_'
    __TEMP_FILE_ENDING: str = '.test.tmp'
    __CONFIG_EXT: str = Config.CONFIG_EXT
    __LOCAL_CONFIG_EXT: str = Config.LOCAL
    __CONFIG_PATH: str = SimulationConfig.CONFIG_PATH
    __SIMULATION_CONFIG: str = SimulationConfig.CONFIG_NAME
    __DATA_DIR: str = os.path.join(os.path.dirname(__file__), 'configs')

    __SELECTED_TESTS: [str] = list()
    """List with names of selected tests to run. If list is empty, all tests will be executed."""
    # __SELECTED_TESTS.append('test_16')
    # __SELECTED_TESTS.append('test_14')

    def __init__(self):
        super().__init__(do_simulation=True, do_analysis=True)
        # self.__id_generator = self.__get_id_generator()
        self.__tests: [str] = list()

    def _setup_config(self) -> dict:
        # test set for config setup
        self.__rename_local_config()
        configs: dict = dict()
        if not self.__SELECTED_TESTS:
            self.__add('test_0', ConfigParser(), configs)
        for file in os.listdir(self.__DATA_DIR):
            if file.endswith(self.__CONFIG_EXT):
                config: ConfigParser = ConfigParser()
                config.read(os.path.join(self.__DATA_DIR, file))
                test_id: str = file.split('.')[-2]
                if self.__SELECTED_TESTS:
                    if test_id not in self.__SELECTED_TESTS:
                        continue
                self.__add(test_id, config, configs)
        return configs

    def _analysis_config(self) -> ConfigParser:
        config_generator: AnalysisConfigGenerator = AnalysisConfigGenerator()
        config_generator.print_results(False)
        config_generator.do_plotting(False)
        config_generator.merge_analysis(False)
        return config_generator.get_config()

    def __add(self, test_id: str, config: ConfigParser, configs: dict) -> None:
        # if test_id is None:
        #     test_id: str = next(self.__id_generator)
        self.__tests.append(test_id)
        configs[test_id] = config

    def __get_id_generator(self) -> str:
        test_id: int = 1
        while True:
            yield self.__TEST_NAME + str(test_id)
            test_id += 1

    def __rename_local_config(self):
        self.__rename_local_simulation_config(self.__LOCAL_CONFIG_EXT, self.__TEMP_FILE_ENDING)

    def __reverse_local_config(self):
        self.__rename_local_simulation_config(self.__TEMP_FILE_ENDING, self.__LOCAL_CONFIG_EXT)

    def __rename_local_simulation_config(self, src_ext: str, dest_ext: str) -> None:
        # 1) Only the simulation.local.ini needs to be renamed.
        # 2) It is safer to copy a file instead of moving or renaming it.
        # 3) Only one place for logic.
        file: str = os.path.join(self.__CONFIG_PATH, self.__SIMULATION_CONFIG + src_ext)
        new_file: str = file.replace(src_ext, dest_ext)
        if os.path.exists(file):
            shutil.copy(file, new_file)
            remove_file(file)
        # for file in os.listdir(self.__CONFIG_PATH):
        #     if file.endswith(self.__LOCAL_CONFIG_EXT):
        #         new_file = file.replace(self.__LOCAL_CONFIG_EXT, self.__TEMP_FILE_ENDING)
        #         os.rename(os.path.join(self.__CONFIG_PATH, file), os.path.join(self.__CONFIG_PATH, new_file))

    # def __rename_local_configs_reverse(self) -> None:
    #     for file in os.listdir(self.__CONFIG_PATH):
    #         if file.endswith(self.__TEMP_FILE_ENDING):
    #             new_file = file.replace(self.__TEMP_FILE_ENDING, self.__LOCAL_CONFIG_EXT)
    #             os.rename(os.path.join(self.__CONFIG_PATH, file), os.path.join(self.__CONFIG_PATH, new_file))

    def clean_up(self) -> None:
        self.__reverse_local_config()
        for test in self.__tests:
            while True:
                try:
                    shutil.rmtree(os.path.join(self.get_results_path(), test))
                    break
                except FileNotFoundError:
                    break
                except PermissionError:
                    print('Waiting for simulation to finish...')
                    time.sleep(1)


if __name__ == "__main__":
    batch_processing: BatchProcessing = SystemTest()
    try:
        batch_processing.run()
    finally:
        batch_processing.clean_up()
        print('System tests Finished')



from configparser import ConfigParser
import ast
from simses.commons.config.simulation.simulation_config import SimulationConfig, create_dict_from, create_list_from, \
    clean_split


class StorageSystemConfig(SimulationConfig):
    """
    Storage system specific configs
    """

    AC_SYSTEM_NAME: int = 0
    AC_SYSTEM_POWER: int = 1
    AC_SYSTEM_DC_VOLTAGE: int = 2
    AC_SYSTEM_CONVERTER: int = 3
    AC_SYSTEM_HOUSING: int = 4
    AC_SYSTEM_HVAC: int = 5

    DC_SYSTEM_NAME: int = 0
    DC_SYSTEM_CONVERTER: int = 1
    DC_SYSTEM_STORAGE: int = 2

    ACDC_CONVERTER_TYPE: int = 0
    ACDC_CONVERTER_NUMBERS: int = 1
    ACDC_CONVERTER_SWITCH: int = 2
    ACDC_CONVERTER_EFFICIENCY: int = 3

    DCDC_CONVERTER_TYPE: int = 0
    DCDC_CONVERTER_POWER: int = 1
    DCDC_CONVERTER_EFFICIENCY: int = 2

    DC_POWER_DISTRIBUTOR_TYPE: int = 0

    HOUSING_TYPE: int = 0
    HOUSING_HIGH_CUBE: int = 1  # Determines whether container to be created with high cube or standard dimensions
    HOUSING_AZIMUTH: int = 2
    HOUSING_ABSORPTIVITY = 3
    HOUSING_GROUND_ALBEDO = 4

    HVAC_TYPE: int = 0
    HVAC_POWER: int = 1
    HVAC_TEMPERATURE_SETPOINT: int = 2
    HVAC_KP_COEFFICIENT: int = 3
    HVAC_KI_COEFFICIENT: int = 4
    HVAC_KD_COEFFICIENT: int = 5

    AMBIENT_TEMPERATURE_TYPE: int = 0
    AMBIENT_TEMPERATURE_CONSTANT: int = 1

    SOLAR_IRRADIATION_TYPE: int = 0

    STORAGE_CAPACITY: int = 0
    STORAGE_TYPE: int = 1

    BATTERY_CELL: int = 2
    BATTERY_SOC: int = 3
    BATTERY_SOH: int = 4

    REDOX_FLOW_STACK: int = 2
    STACK_MODULE_POWER: int = 3
    REDOX_FLOW_PUMP_ALGORITHM: int = 4

    FUEL_CELL_TYPE: int = 2
    FUEL_CELL_POWER: int = 3
    ELECTROLYZER_TYPE: int = 4
    ELECTROLYZER_POWER: int = 5
    HYDROGEN_STORAGE: int = 6
    HYDROGEN_TANK_PRESSURE: int = 7

    SECTION: str = 'STORAGE_SYSTEM'

    STORAGE_SYSTEM_DC: str = 'STORAGE_SYSTEM_DC'
    STORAGE_SYSTEM_AC: str = 'STORAGE_SYSTEM_AC'
    ACDC_CONVERTER: str = 'ACDC_CONVERTER'
    DCDC_CONVERTER: str = 'DCDC_CONVERTER'
    HVAC: str = 'HVAC'
    HOUSING: str = 'HOUSING'
    STORAGE_TECHNOLOGY: str = 'STORAGE_TECHNOLOGY'
    AMBIENT_TEMPERATURE_MODEL: str = 'AMBIENT_TEMPERATURE_MODEL'
    SOLAR_IRRADIATION_MODEL: str = 'SOLAR_IRRADIATION_MODEL'
    THERMAL_SIMULATION: str = 'THERMAL_SIMULATION'
    CYCLE_DETECTOR: str = 'CYCLE_DETECTOR'
    POWER_DISTRIBUTOR_DC: str = 'POWER_DISTRIBUTOR_DC'
    POWER_DISTRIBUTOR_AC: str = 'POWER_DISTRIBUTOR_AC'
    LINEAR_DISTRIBUTION_Z_VALUE: str = 'LINEAR_DISTRIBUTION_Z_VALUE'
    LINEAR_OPTION: str = 'LINEAR_OPTION'

    def __init__(self, config: ConfigParser, path: str = None):
        super().__init__(path, config)

    @property
    def storage_systems_dc(self) -> [[str]]:
        """Returns a list of dc storage systems"""
        props: [str] = clean_split(self.get_property(self.SECTION, self.STORAGE_SYSTEM_DC))
        return create_list_from(props)

    @property
    def storage_systems_ac(self) -> [[str]]:
        """Returns a list of ac storage systems"""
        props: [str] = clean_split(self.get_property(self.SECTION, self.STORAGE_SYSTEM_AC))
        return create_list_from(props)

    @property
    def acdc_converter(self) -> dict:
        """Returns a list of acdc converter"""
        props: [str] = clean_split(self.get_property(self.SECTION, self.ACDC_CONVERTER))
        return create_dict_from(props)

    @property
    def dcdc_converter(self) -> dict:
        """Returns a list of acdc converter"""
        props: [str] = clean_split(self.get_property(self.SECTION, self.DCDC_CONVERTER))
        return create_dict_from(props)

    @property
    def hvac(self) -> dict:
        """Returns a list of hvac systems"""
        props: [str] = clean_split(self.get_property(self.SECTION, self.HVAC))
        return create_dict_from(props)

    @property
    def housing(self) -> dict:
        """Returns a list of housing objects"""
        props: [str] = clean_split(self.get_property(self.SECTION, self.HOUSING))
        return create_dict_from(props)

    @property
    def storage_technologies(self) -> dict:
        """Returns a list of storage technologies"""
        props: [str] = clean_split(self.get_property(self.SECTION, self.STORAGE_TECHNOLOGY))
        return create_dict_from(props)

    @property
    def ambient_temperature_model(self) -> list:
        """Returns name of ambient temperature model"""
        props: [str] = clean_split(self.get_property(self.SECTION, self.AMBIENT_TEMPERATURE_MODEL), ',')
        return props

    @property
    def solar_irradiation_model(self) -> list:
        """Returns name of solar irradiation model"""
        props: [str] = clean_split(self.get_property(self.SECTION, self.SOLAR_IRRADIATION_MODEL), ',')
        return props

    @property
    def thermal_simulation(self) -> bool:
        """Returns user preference to enable/disable thermal simulation"""
        props: bool = ast.literal_eval(clean_split(self.get_property(self.SECTION, self.THERMAL_SIMULATION), ',')[0])
        return props

    @property
    def cycle_detector(self) -> str:
        """Returns name of cycle detector"""
        return self.get_property(self.SECTION, self.CYCLE_DETECTOR)

    @property
    def power_distributor_dc(self) -> [str]:
        """Returns name of cycle detector"""
        return clean_split(self.get_property(self.SECTION, self.POWER_DISTRIBUTOR_DC), ',')

    @property
    def power_distributor_ac(self) -> str:
        """Returns name of cycle detector"""
        return self.get_property(self.SECTION, self.POWER_DISTRIBUTOR_AC)

    @property
    def linear_distributor_z_value(self) -> str:
        """Returns name of cycle detector"""
        return self.get_property(self.SECTION, self.LINEAR_DISTRIBUTION_Z_VALUE)

    @property
    def linear_option(self) -> str:
        """Returns name of cycle detector"""
        return self.get_property(self.SECTION, self.LINEAR_OPTION)



import numpy as np


class Hydrogen:

    FARADAY_CONST: float = 96485.3321  # As/mol
    IDEAL_GAS_CONST: float = 8.314462  # J/(mol K)
    REAL_GAS_FACTOR = 1.0006  # from: "Wasserstoff in der Fahrzeugtechnik" under normal conditions: p = 1 bar, T = 0°C
    ISENTROP_EXPONENT = 1.4098  # # from: "Wasserstoff in der Fahrzeugtechnik" under normal conditions: p = 1 bar, T = 0°C
    MOLAR_MASS_HYDROGEN = 1.00794 * 10 ** (-3) # kg/mol
    MOLAR_MASS_WATER: float = 0.018015  # kg/mol
    MOLAR_MASS_OXYGEN = 15.999 * 10 ** (-3)  # kg/mol
    HEAT_CAPACITY_WATER: float = 4184  # J/(kg K)
    HEAT_CAPACITY_HYDROGEN = 14304  # J/(kg K)
    HEAT_CAPACITY_OXYGEN = 920  # J/(kg K)
    NORM_QUBIC_M = 11.1235  # Nm³/kg
    DENSITY_WATER = 1000  # kg/m³
    EPS = np.finfo(float).eps
    LOWER_HEATING_VALUE = 33.327  # kWh/kg lower heating value H2
    VAN_D_WAALS_COEF_A = 0.02452065  # van der Waals coefficient A of H2 (J*m³)/mol²
    VAN_D_WAALS_COEF_B = 2.65e-5  # van der Waals coefficient B of H2  m³/mol
    BOLTZ_CONST = 1.38e-23  # J/K

    @staticmethod
    def realgas_factor(p_1, p_2):
        """ calculates real gas factor for hydrogen based on grafig out of "Wasserstoff in der Fahrzeugtechnik" p. 49
        for T = 300 K """
        slope = 8.5 * 10 ** (-4)
        return 1/2 * (2 + slope * (p_1 + p_2))

    @staticmethod
    def isentropic_exponent(p_1, p_2):
        """ calculates mean isentropic exponent for hydrogen for compression from pressure p_1 to pressure p_2 in bar
        based on diagram in Thermodynamic of Pressurized Gas Strorage in Hydrogen Science and Engineering (2016)
        valid for T = 50 °C """
        slope = 9.047 * 10 ** (-4)
        return 1/2 * (2 * 1.404 + slope * (p_1 + p_2))



import pandas as pd

from simses.commons.config.data.fuel_cell import FuelCellDataConfig
from simses.commons.constants import Hydrogen
from simses.commons.log import Logger
from simses.technology.hydrogen.fuel_cell.stack.stack_model import \
    FuelCellStackModel


class PemFuelCell(FuelCellStackModel):
    """An PEM-Fuelcell is a special typ of a Fuelcell"""


    __POWER_HEADER = 'cellpower at 1 bar'
    __VOLTAGE_IDX = 1
    __POWER_IDX = 1
    __CURRENT_IDX = 0

    def __init__(self, fuel_cell_maximal_power: float, fuel_cell_data_config: FuelCellDataConfig):
        super().__init__()
        self.__log: Logger = Logger(type(self).__name__)
        self.__SPEC_MAX_POWER = 0.62  # W/cm2  TODO (JK) evaluate value
        self.__NOM_STACK_POWER = fuel_cell_maximal_power
        self.__GEOM_AREA = fuel_cell_maximal_power / self.__SPEC_MAX_POWER  # cm2
        self.__NUMBER_CELLS = 1
        self.__PC_FILE = fuel_cell_data_config.pem_fuel_cell_pc_file
        self.__POWER_FILE = fuel_cell_data_config.pem_fuel_cell_power_file
        self.__mass_flow_h2 = 0  # mol/s
        self.__MAX_CURRENT_DENSITY = 5  # A/cm2
        self.__MIN_CURRENT_DENSITY = 0  # mA
        self.__HEAT_CAPACITY = Hydrogen.EPS
        self.__polarization_curve = pd.read_csv(self.__PC_FILE)  # V/(mA/cm2)
        self.__power_curve = pd.read_csv(self.__POWER_FILE)  # W/cm2
        self.__current_arr = self.__polarization_curve.iloc[:, self.__CURRENT_IDX] * self.__GEOM_AREA  # mA
        self.__polarization_arr = self.__polarization_curve.iloc[:, self.__VOLTAGE_IDX]  # V
        self.__power_arr = self.__power_curve.iloc[:, self.__POWER_IDX] * self.__GEOM_AREA  # W

    def calculate(self, power: float):
        power_abs = abs(power)
        power_idx = abs(self.__power_curve[self.__POWER_HEADER] * self.__GEOM_AREA - power_abs).idxmin()   # hier wird eine positive Leistung erwartet!!!
        self.__voltage = self.__polarization_curve.iloc[power_idx, self.__VOLTAGE_IDX]  # V
        self.__current = - self.__power_curve.iloc[power_idx, self.__CURRENT_IDX] * self.__GEOM_AREA  # mA
        self.__hydrogen_consumption = self.__current / (2 * Hydrogen.FARADAY_CONST)  # mol/s

    def get_current(self):
        #print('the current is: ' + str(self.__current) + ' mA')
        return self.__current

    def get_hydrogen_consumption(self):
        #print('the massflow of hydrogen ist: ' + str(self.__mass_flow_h2) + ' mol/s')
        return - self.__hydrogen_consumption  # positive value

    def get_voltage(self):
        #print('the voltage is: ' + str(self.__voltage) + ' V')
        return self.__voltage

    def get_nominal_stack_power(self):
        return self.__NOM_STACK_POWER

    def get_number_cells(self):
        return self.__NUMBER_CELLS

    def get_geom_area_stack(self):
        return self.__GEOM_AREA

    def get_heat_capactiy_stack(self):
        return self.__HEAT_CAPACITY

    def close(self):
        self.__log.close()

 #p = PemFuelCell()
 #p.calculate(100)
 #p.get_current()
 #p.get_hydrogen_consumption()
 #p.get_voltage()




from simses.commons.state.technology.storage import StorageTechnologyState


class HydrogenState(StorageTechnologyState):
    """
    Current physical state of the hydrogen storage components with the main electrical parameters.
    """

    SYSTEM_AC_ID: str = 'StorageSystemAC'
    SYSTEM_DC_ID: str = 'StorageSystemDC'
    SOC: str = 'SOC in p.u.'
    SOH = 'State of health in p.u.'
    POWER: str = 'Power in W'
    VOLTAGE: str = 'voltage of hydrogen system in V'
    CURRENT: str = 'current of hydrogen storage system in A'
    POWER_LOSS: str = 'Power loss in W'
    CAPACITY: str = 'capacity in Wh'
    FULFILLMENT: str = 'fulfillment in p.u.'
    TEMPERATURE: str = 'temperature in K'
    MAX_CHARGE_POWER = 'Maximum charging power in W'
    MAX_DISCHARGE_POWER = 'Maximum discharging power in W'

    def __init__(self, system_id: int, storage_id: int):
        super().__init__()
        self._initialize()
        self.set(self.SYSTEM_AC_ID, system_id)
        self.set(self.SYSTEM_DC_ID, storage_id)

    @property
    def is_charge(self) -> bool:
        return self.power > 0

    @property
    def voltage(self) -> float:
        return self.get(self.VOLTAGE)

    @voltage.setter
    def voltage(self, value: float) -> None:
        self.set(self.VOLTAGE, value)

    @property
    def power(self) -> float:
        return self.get(self.POWER)

    @power.setter
    def power(self, value: float) -> None:
        self.set(self.POWER, value)

    @property
    def current(self) -> float:
        return self.get(self.CURRENT)

    @current.setter
    def current(self, value: float) -> None:
        self.set(self.CURRENT, value)

    @property
    def power_loss(self) -> float:
        return self.get(self.POWER_LOSS)

    @power_loss.setter
    def power_loss(self, value: float) -> None:
        self.set(self.POWER_LOSS, value)

    @property
    def soc(self) -> float:
        return self.get(self.SOC)

    @soc.setter
    def soc(self, value: float) -> None:
        self.set(self.SOC, value)

    @property
    def soh(self) -> float:
        return self.get(self.SOH)

    @soh.setter
    def soh(self, value: float) -> None:
        self.set(self.SOH, value)

    @property
    def capacity(self) -> float:
        return self.get(self.CAPACITY)

    @capacity.setter
    def capacity(self, value: float):
        self.set(self.CAPACITY, value)

    @property
    def fulfillment(self) -> float:
        return self.get(self.FULFILLMENT)

    @fulfillment.setter
    def fulfillment(self, value: float):
        self.set(self.FULFILLMENT, value)

    @property
    def temperature(self) -> float:
        return self.get(self.TEMPERATURE)

    @temperature.setter
    def temperature(self, value: float) -> None:
        self.set(self.TEMPERATURE, value)

    @property
    def id(self) -> str:
        return 'HYDROGEN' + str(self.get(self.SYSTEM_AC_ID)) + str(self.get(self.SYSTEM_DC_ID))

    @classmethod
    def sum_parallel(cls, states: []):
        pass

    @classmethod
    def sum_serial(cls, states: []):
        pass

    @property
    def max_charge_power(self) -> float:
        return self.get(self.MAX_CHARGE_POWER)

    @max_charge_power.setter
    def max_charge_power(self, value: float) -> None:
        self.set(self.MAX_CHARGE_POWER, value)

    @property
    def max_discharge_power(self) -> float:
        return self.get(self.MAX_DISCHARGE_POWER)

    @max_discharge_power.setter
    def max_discharge_power(self, value: float) -> None:
        self.set(self.MAX_DISCHARGE_POWER, value)



from simses.commons.profile.power.generation import GenerationProfile
from simses.commons.profile.power.power_profile import PowerProfile
from simses.commons.state.energy_management import EnergyManagementState
from simses.commons.state.system import SystemState
from simses.logic.energy_management.strategy.operation_priority import OperationPriority
from simses.logic.energy_management.strategy.operation_strategy import OperationStrategy


class ResidentialPvGreedy(OperationStrategy):
    """
    Operation Strategy for a residential home storage system in combination with PV generation.
    The algorithm fills the BESS as fast as possible without consideration for the grid by meeting the residual load
    at all times.
    """

    def __init__(self, power_profile: PowerProfile, pv_generation_profile: GenerationProfile):
        super().__init__(OperationPriority.MEDIUM)
        self.__load_profile: PowerProfile = power_profile
        self.__pv_profile: GenerationProfile = pv_generation_profile
        self.__pv_power = 0
        self.__load_power = 0

    def next(self, time: float, system_state: SystemState, power: float = 0) -> float:
        self.__load_power = self.__load_profile.next(time)
        self.__pv_power = self.__pv_profile.next(time)
        residual_charge_power = self.__pv_power - self.__load_power
        return residual_charge_power

    def update(self, energy_management_state: EnergyManagementState) -> None:
        energy_management_state.pv_power = self.__pv_power
        energy_management_state.load_power = self.__load_power

    def clear(self) -> None:
        self.__pv_power = 0.0
        self.__load_power = 0.0

    def close(self) -> None:
        self.__load_profile.close()
        self.__pv_profile.close()



import numpy as np
import matplotlib.pyplot as plt
import math as m
import pandas as pd


# physical constants  / physikalische Konstanten
IDEAL_GAS_CONST: float = 8.314462  # J/(mol K)
FARADAY_CONST: float = 96485.3321  # As/mol
ATM = 1.0  # bar

# fuel cell parameters / Brennstoffzellen Parameter
TEMPERATURE_REF_K: float = 298.15  # K
TEMPERATURE_C_in_K: float = 273.15  # K
RESISTANCE = 0.144787415  # Ohm*cm2

AREA_CELL = 150.0  # cm2
Max_CURRENT = 100.0  # A
Max_CURRENT_Density = Max_CURRENT/AREA_CELL  # A/cm2
REV_VOLTAGE_STC = 1.229  # V

PRESSURE_ANODE = 1.36  # pressure anode in bar
PRESSURE_CATHODE = 1.0  # pressure cathode in bar
PH2 = PRESSURE_ANODE / ATM  # bar to atm
PO2 = PRESSURE_CATHODE * 0.21 / ATM  # bar to atm

B = 0.044  # Konzentrationskoeffizient
N = 70  # Anzahl Zellen
alpha = 0.207900099  # Ladungsübertragungskoeffizient
i_0_ref = 1.89702*10**-4  # Austauschstomdichte [A/cm2] bei einer Referenztemperatur
E_A = 40000  # Aktivierungsenergie (wichtig für Austauschstromdichte [J/mol]
i_n = 0.0100310306  # internal current density [A/cm2]

# Auxiliary Arrays / Hilfs-Arrays
current_density = np.linspace(0, Max_CURRENT_Density, 1001)
cell_voltage = np.ones(len(current_density))
stack_voltage = np.ones(len(current_density))
cell_power = np.ones(len(current_density))
stack_power = np.ones(len(current_density))
cell_temperatur = np.ones(len(current_density))


def T_cell_Calc():
    """
    Calculate T_cell.
    :param i: cell current [A/cm2]
    :type i: float
    :return: T as float
    """
    try:
        result = (0.53 * current_density[i] * AREA_CELL + 26.01 + TEMPERATURE_C_in_K)
        return result
    except (TypeError, ZeroDivisionError):
        return None

def B1_Calc(n=2):
    """
    Calculate B (Constant in the mass transfer term).
    :param n: number of moles of electrons transferred in the balanced equation occurring in the fuel cell
    :type n: int
    :return: B1 as float
    """
    try:
        T = cell_temperatur[i]
        result = (IDEAL_GAS_CONST * T) / (n * FARADAY_CONST)
        return result
    except (TypeError, ZeroDivisionError):
        return None


# Nernst Voltage
def Enernst_Calc(PH2, PO2):
    """
    Calculate Enernst.
    :param PH2: partial pressure [atm]
    :type PH2 : float
    :param PO2: partial Pressure [atm]
    :type PO2: float
    :return: Enernst [V} as float
    """
    try:
        B1 = B1_Calc()  # (R * Tcell) / (2*F)
        T = cell_temperatur[i]

        result = REV_VOLTAGE_STC + B1 * m.log(PH2 * PO2 ** 0.5) - 0.85 * 10 ** -3 * (T - TEMPERATURE_REF_K)
        return result
    except (TypeError, OverflowError, ValueError):
        print(
            "[Error] Enernst Calculation Failed (PH2:%s, PO2:%s)" %
            (str(PH2), str(PO2)))


def i_0_Calc_from_temperature():
    """
    :param i_0_ref: Referenz Austauschstromdichte in [A/cm2]
    :type i_0_ref: float
    :param i: cell load current [A/cm2]
    :type i: float
    :return: i_0 Austauschstromdichte in Abhängigkeit der Temperatur  in [A/cm2]
    """
    Temperature = cell_temperatur[i]
    i_0 = i_0_ref * m.exp((-E_A/IDEAL_GAS_CONST)*((1/Temperature)-(1/TEMPERATURE_REF_K)))
    return i_0


def Eta_Conc_Calc(i, B, imax):
    """
    Calculate Eta concentration.
    :param i: cell load current [A/cm2]
    :type i :float
    :param imax: maximal cell current [A/cm2]
    :type imax: float
    :return: Eta concentration [V] as float
    """
    try:
        if i >= 0 and i + i_n < imax:
            result = - B * m.log(1 - ((i + i_n) / imax))
            return result

    except (TypeError, ZeroDivisionError, OverflowError, ValueError):
        print(
            "[Error] Eta Concentration Calculation Failed (i:%s, B1:%s, imax:%s)" %
            (str(i), str(B), str(imax)))


def Eta_Ohmic_Calc(R_total):
    """
    Calculate Eta ohmic.
    :param i: cell current [A/cm2]
    :type i: float
    :param R_total: Gesamtwiderstand einer Zelle [Ohm]
    :type R_total: float

    """
    try:
        if i != -i_n:
            result = (current_density[i] + i_n) * R_total
            return result
        return 0
    except (TypeError, ZeroDivisionError):
        print(
            "[Error] Eta Ohmic Calculation Failed (i:%s, R_total:%s)" %
            (str(i), str(R_total)))


def Eta_Act_Calc(i, alpha, i_n):
    """
    Calculate Eta activation.
    :param alpha: Ladungsübertragungskoeffizient
    :type alpha: float
    :param i: cell current [A/cm2]
    :type i: float
    :param i_n: interner Zellstrom [A/cm2]
    :type i_n: float
    :param i_0: Austauschstromdichte [A/cm2]
    :type i_0: float
    :return:  Eta activation [V] as float
    """
    try:
        i_0 = i_0_Calc_from_temperature()
        B1 = B1_Calc()  # (R * Tcell) / (2*F)

        result = (B1 / alpha) * (m.log((i + i_n) / i_0))
        return result

    except (TypeError, OverflowError, ValueError):
        print(
            "[Error] Eta Activation Calculation Failed (i:%s, alpha:%s, i_n:%s, i_0:%s)" %
            (str(i), str(alpha), str(i_n), str(i_0)))


def VStack_Calc(N, Vcell):
    """
    Calculate VStack.
    :param N: number of single cells
    :type N: int
    :param Vcell: cell voltage [V}
    :type Vcell: float
    :return: VStack [V] as float
    """
    try:
        result = N * Vcell
        return result
    except TypeError:
        print(
            "[Error] VStack Calculation Error (N:%s, Vcell:%s)" %
            (str(N), str(Vcell)))


def Loss_Calc(Eta_Act, Eta_Ohmic, Eta_Conc):
    """
    Calculate loss.
    :param Eta_Act: Eta activation [V]
    :type Eta_Act : float
    :param Eta_Ohmic: Eta ohmic [V]
    :type Eta_Ohmic : float
    :param Eta_Conc: Eta concentration [V]
    :type Eta_Conc : float
    :return: loss [V] as float
    """
    try:
        result = Eta_Act + Eta_Ohmic + Eta_Conc
        return result
    except TypeError:
        print(
            "[Error] Loss Calculation Error (Eta_Act:%s, Eta_Ohmic:%s, Eta_Conc:%s)" %
            (str(Eta_Act), str(Eta_Ohmic), str(Eta_Conc)))


def Vcell_Calc(Enernst, Loss):
    """
    Calculate cell voltage.
    :param Enernst:  Nernstvoltage [V}
    :type Enernst : float
    :param Loss:  loss [V]
    :type Loss : float
    :return:  cell voltage [V] as float
    """
    try:
        result = Enernst - Loss
        return result
    except TypeError:
        print(
            "[Error] Vcell Calculation Error (Enernst:%s, Loss:%s)" %
            (str(Enernst), str(Loss)))


def Power_Calc(Vcell, i):
    """
    Calculate power.
    :param Vcell: Vcell Voltage [V]
    :type Vcell : float
    :param i: cell current [A/cm2]
    :type i : float
    :return: cell power [W/cm2] as float
    """
    try:
        result = Vcell * i
        return result
    except TypeError:
        print(
            "[Error] Power Calculation Error (Vcell:%s, i:%s)" %
            (str(Vcell), str(i)))


def PowerStack_Calc(Power, N):
    """
    Calculate power_stack.
    :param Power: single cell power [W/cm2]
    :type Power : float
    :param N: number of single cells
    :type N : int
    :return: power stack [W] as float
    """
    try:
        result = N * Power * AREA_CELL
        return result
    except TypeError:
        print(
            "[Error] Power Stack Calculation Error (Power:%s, N:%s)" %
            (str(Power), str(N)))


for i in range(0, len(current_density)):
    a = current_density[i]
    cell_temperatur[i] = T_cell_Calc()
    Enernst = Enernst_Calc(PH2, PO2)
    Eta_Act = Eta_Act_Calc(a, alpha, i_n)
    Eta_Ohmic = Eta_Ohmic_Calc(RESISTANCE)
    Eta_Conc = Eta_Conc_Calc(a, B, Max_CURRENT_Density)
    Loss = Loss_Calc(Eta_Act, Eta_Ohmic, Eta_Conc)
    cell_voltage[i] = Vcell_Calc(Enernst, Loss)
    stack_voltage[i] = Vcell_Calc(N, cell_voltage[i])
    cell_power[i] = Power_Calc(cell_voltage[i], a)
    stack_power[i] = PowerStack_Calc(N, cell_power[i])

x = current_density * 150
y1 = cell_voltage * 70
y2 = cell_power * 150 * 70


#########################
#  Messungen dargestellt:
current_messung1 = (13.2, 16.08, 33.2, 53.2, 72.5, 76.7)  # neue Messungen scheinbar besser =)
voltage_messung1 = (57.09, 56.18, 52.27, 48.50, 44.65, 42.67)  # neue Messungen scheinbar besser =) --> siehe Excel File

current_messung2 = (12.29, 16.78, 33.68, 53.96, 73.52, 78.15)  # neue Messungen scheinbar besser =)
voltage_messung2 = (57.09, 56.18, 52.27, 48.43, 44.62, 42.68)  # neue Messungen scheinbar besser =) --> siehe Excel File

current_messung3 = (12.96, 18.13, 35.20, 55.81, 75.13, 79.94)  # neue Messungen scheinbar besser =)
voltage_messung3 = (57.07, 56.16, 52.20, 48.42, 44.62, 42.67)  # neue Messungen scheinbar besser =) --> siehe Excel File

#########################


fig, axs = plt.subplots(2, 1, constrained_layout=True)
axs[0].plot(x, y1)

#  Messungen plotten ######
axs[0].plot(current_messung1, voltage_messung1)
axs[0].plot(current_messung2, voltage_messung2)
axs[0].plot(current_messung3, voltage_messung3)
###########################

axs[0].set_title('U-I-Kennlinie')
axs[0].set_xlabel('Stromdichte (A/cm2)')
axs[0].set_ylabel('Zellspannung (V)')
fig.suptitle('Zellspannung und Zellleistung in Abhängigkeit der Stromdichte', fontsize=12)

axs[1].plot(x, y2)
axs[1].set_xlabel('Stromdichte (A/cm2)')
axs[1].set_title('P-I-Kennlinie')
axs[1].set_ylabel('Zellleistung (W/cm2)')

plt.show()

dict = {'currentdensity': current_density, 'cellvoltage at 1 bar': cell_voltage}
dict2 = {'currentdenstiy': current_density, 'cellpower at 1 bar': cell_power}

df = pd.DataFrame(dict)
df2 = pd.DataFrame(dict2)

#df.to_csv('polarizationcurve_jupiter.csv', index=False)
#df2.to_csv('powercurve_jupiter.csv', index=False)



import math
from simses.commons.log import Logger
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.profile import ProfileConfig
from simses.commons.profile.power.generation import GenerationProfile
from simses.commons.state.energy_management import EnergyManagementState
from simses.commons.state.system import SystemState
from simses.logic.energy_management.strategy.operation_priority import OperationPriority
from simses.logic.energy_management.strategy.operation_strategy import OperationStrategy


class ResidentialPvFeedInDamp(OperationStrategy):
    """
    Operation Strategy for a residential home storage system in combination with PV generation.
    The algorithm plans the charging of the lithium_ion according to a PV prediction. It tries to provide a fully charged
    BESS at sundown.
    """

    def __init__(self, power_profile, general_config: GeneralSimulationConfig, profile_config: ProfileConfig):
        super().__init__(OperationPriority.MEDIUM)
        self.__log: Logger = Logger(type(self).__name__)
        self.__start = general_config.start
        self.__end = general_config.end
        self.__timestep = general_config.timestep
        self.__load_profile = power_profile
        self.__pv_profile = GenerationProfile(profile_config, general_config)
        self.__p_pv_peak = 1
        self.__Wh_to_Ws = 3600
        self.__s_per_day = 86400
        self.__steps_per_day = int(self.__s_per_day / self.__timestep)
        self.__pv_profile_copy_for_prediction = GenerationProfile(profile_config, general_config)
        self.__last_pv_generation: list = self.pv_prediction(self.__pv_profile_copy_for_prediction)
        self.__pv_power = 0
        self.__load_power = 0

        self.__eta = 0.95

    def next(self, time: float, system_state: SystemState, power: float = 0) -> float:
        self.__pv_power = self.__pv_profile.next(time)
        self.__load_power = self.__load_profile.next(time)

        residual_load = self.__pv_power - self.__load_power
        if residual_load > 0 and self.__pv_power > 0:
            day = math.floor(
                (time - self.__start) / self.__s_per_day)
            capacity_remaining = system_state.capacity * (1 - system_state.soc)

            energy_remaining = capacity_remaining * self.__Wh_to_Ws
            remaining_time = self.__last_pv_generation[
                                   day] - time - 3600  # An hour before last pv production the storage should be filled
            if remaining_time > 0:  # Preventing division by 0
                power_ref = energy_remaining / (remaining_time * self.__eta)  # Ws
            else:
                power_ref = energy_remaining / (1 * self.__eta)  # Ws
            battery_power = min(power_ref, residual_load)
        else:
            battery_power = residual_load
        return battery_power

    def pv_prediction(self, pv_generation_profile) -> list:
        """
        Function to determine the last pv production of every day of the simulation
        and storing the timestamps into a list.

        Returns
        -------
        list:
            timestamps of the last pv production of every day of the simulation.
        """
        self.__log.info('PV Prediction is performed this may take some time.')
        pv_profile: list = pv_generation_profile.profile_data_to_list()
        pv = pv_profile.copy()

        timesteps = (self.__end - self.__start) / self.__timestep
        days = int(timesteps / self.__steps_per_day) + 1

        last_pv_generation = []
        for day in range(days):
            if day == days - 1:  # Last day special consideration (not full day)
                beginning_of_day = day * self.__steps_per_day * self.__timestep
                remaining_steps = int(
                    (self.__end - self.__start - beginning_of_day) / self.__timestep)
                for step in range(remaining_steps):
                    current_step = day * self.__steps_per_day + remaining_steps - step
                    if pv[current_step - 1] > 0:
                        last_pv_generation.append(self.__start + current_step * self.__timestep)
                        break
                    if step == remaining_steps - 1:  # No Generation on this day
                        last_pv_generation.append(None)
            else:
                for step in range(self.__steps_per_day):
                    current_step = (day + 1) * self.__steps_per_day - step
                    if pv[current_step] > 0:
                        last_pv_generation.append(self.__start + current_step * self.__timestep)
                        break
                    if step == self.__steps_per_day - 1:  # No Generation on this day
                        last_pv_generation.append(None)
        self.__log.info('PV Prediction done.')
        return last_pv_generation

    def update(self, energy_management_state: EnergyManagementState) -> None:
        energy_management_state.pv_power = self.__pv_power
        energy_management_state.load_power = self.__load_power

    def clear(self) -> None:
        self.__pv_power = 0.0
        self.__load_power = 0.0

    def close(self) -> None:
        self.__load_profile.close()
        self.__pv_profile.close()



from simses.system.thermal.solar_irradiation.solar_irradiation_model import SolarIrradiationModel


class NoSolarIrradiationModel(SolarIrradiationModel):
    """
        NoSolarIrradiationModel returns a value of 0.0 for the thermal load due to the solar irradiation
        """

    def __init__(self):
        super().__init__()

    def get_heat_load(self, time_stamp) -> float:
        return 0.0  # Not applicable

    def get_global_horizontal_irradiance(self,time_step) -> float:
        return 0.0  # Not applicable

    def create_instance(self) -> SolarIrradiationModel:
        """
        reinstantiates the SolarIrradiationModel
        :return: SolarIrradiationModel
        """
        return NoSolarIrradiationModel()

    def close(self):
        pass



from math import ceil, floor
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from simses.analysis.evaluation.plotting.axis import Axis
from simses.analysis.evaluation.plotting.plotter import Plotting
from simses.analysis.evaluation.plotting.sankey_diagram import SankeyDiagram
from simses.analysis.evaluation.plotting.sunburst_diagram import SunburstDiagram


class PlotlyPlotting(Plotting):

    static_figure_index = 1
    __nbins = 100

    class Linestyle:
        DOTTED: str = 'dot'
        SOLID: str = 'solid'
        DASHED: str = 'dash'
        DASH_DOT: str = 'dashdot'

    def __init__(self, title: str, path: str):
        super().__init__()
        self.__title = title
        self.__path = path
        self.__figures: list = list()

    def layout(self, fig, xaxis: Axis=None, yaxis: [Axis]=None, hist: bool=False):

        fig.layout.title = self.__title
        fig.layout.font = dict(
            family="Times New Roman, monospace",  # TUM Guideline: Helvetica
            size=18,
            # color="#7f7f7f"
        )

        if xaxis and yaxis:
            if len(yaxis) == 1:
                if hist:
                    yaxistitle = 'Relative frequency'
                else:
                    yaxistitle = yaxis[0].label
            else:
                yaxistitle = ""

            fig.update_layout(
                xaxis_title=xaxis.label,
                yaxis_title=yaxistitle,
                yaxis=dict(
                    showexponent='all',
                    exponentformat='e'
                ),
                showlegend=True,
                template="none",
            )
            fig.update_xaxes(showgrid=True, gridwidth=2, gridcolor='Lightgray',
                             showline=True, linewidth=2, linecolor='black')
            fig.update_yaxes(showgrid=True, gridwidth=2, gridcolor='Lightgray',
                             showline=True, linewidth=2, linecolor='black')

    def lines(self, xaxis: Axis, yaxis: [Axis], secondary=[]):
        fig = make_subplots(specs=[[{"secondary_y": True}]])
        self.layout(fig, xaxis, yaxis)
        fig.update_layout(legend=dict(x=0,y=1,traceorder="normal"))

        for i in range(len(yaxis)):
            if i in secondary:
                secondary_y_axis = True
            else:
                secondary_y_axis = False
            fig.add_trace(go.Scatter(
                x=xaxis.data,
                y=yaxis[i].data,
                mode='lines',
                name=yaxis[i].label,
                line=dict(color=yaxis[i].color, dash=yaxis[i].linestyle)
                ), secondary_y=secondary_y_axis
            )
        self.show(fig)

    def get_figures(self) -> list:
        return self.__figures

    # @staticmethod
    # def convert_to_html(figure) -> str:
    #     return figure.to_html(auto_play= False,
    #                           include_plotlyjs=True,
    #                           include_mathjax=False,
    #                           #post_script=plot_id,
    #                           full_html=False,
    #                           #default_height=()),
    #                           validate=True
    #                           )

    def show(self, fig):
        self.__figures.append(fig)
        # pio.write_image(fig,
        #                 self.__path+self.alphanumerize(self.__title)+"_{}.svg".format(Plotting.static_figure_index),
        #                 format='svg',#or 'svg'
        #                 scale=None,#>1 increases resolution
        #                 width=1600,
        #                 height=800,
        #                 validate=True
        #                 )
        self.static_figure_index += 1

    def histogram(self, xaxis: Axis, yaxis: [Axis]):
        cols = ceil(len(yaxis) / 2)
        rows = ceil(len(yaxis) / 2)
        fig = make_subplots(rows=rows, cols=cols)
        for x in range(len(yaxis)):
            ydata = yaxis[x]
            fig.append_trace(go.Histogram(
                x=ydata.data,
                histnorm='percent',
                name=ydata.label,
                nbinsx=self.__nbins,
                marker={"color": ydata.color}
            ), row=floor(x/rows) + 1, col=(x % cols) +1)
        self.layout(fig, xaxis=Axis(data=None, label=None), yaxis=yaxis, hist=True)
        self.show(fig)

    def bar(self, yaxis: [Axis], bars: int):
        titles = [yaxis[x].label.split(" ", 1)[1] for x in range(0, len(yaxis), bars)]
        cols = ceil(len(yaxis) / bars) # 2 pro row
        rows = ceil(len(yaxis) / (bars + cols))
        fig = make_subplots(rows=rows, cols=cols, subplot_titles=titles)
        for i in range(0, len(yaxis), bars):
            labels = []
            values = []
            colors = []
            for j in range(bars):
                ydata = yaxis[i + j]
                labels.append(ydata.label.split()[0])
                values.append(ydata.data.sum())
                colors.append(ydata.color)
            fig.append_trace(go.Bar(x=labels, y=values, name=titles[floor(i / bars)], marker_color=colors),
                             row=ceil((i + j) / 4), col=ceil((i + j) / 2))
        self.layout(fig, xaxis=Axis(data=None, label=None), yaxis=yaxis)
        fig.update_layout(showlegend=False)
        self.show(fig)

    def sankey_diagram(self, node_links: dict):
        fig = go.Figure(data=[go.Sankey(
            node=dict(
                pad=15,
                thickness=20,
                line=dict(color="black", width=0.5),
                label=node_links[SankeyDiagram.NODE_LABELS],
                color=node_links[SankeyDiagram.NODE_COLORS]),
            link=dict(
                source=node_links[SankeyDiagram.SOURCES],
                target=node_links[SankeyDiagram.TARGETS],
                value=node_links[SankeyDiagram.VALUES],
                color=node_links[SankeyDiagram.LINK_COLORS]
            ))])
        self.layout(fig)
        self.show(fig)

    def sunburst_diagram(self, parameters: dict):
        fig = go.Figure(go.Sunburst(
            labels=parameters[SunburstDiagram.LABELS],
            parents=parameters[SunburstDiagram.PARENTS],
            values=parameters[SunburstDiagram.VALUES],
            textinfo='percent entry',
            texttemplate = '%{percentRoot:.2%}',
            branchvalues='total',
        ))
        fig.update_layout(margin=dict(t=35, l=0, r=0, b=0))
        self.layout(fig)
        self.show(fig)

    def subplots(self, xaxis: Axis, yaxis: [Axis]):
        cols: int = 2  # ceil(len(yaxis) / 2)
        rows: int = ceil(float(len(yaxis)) / float(cols))
        fig = make_subplots(rows=rows, cols=cols)
        row: float = 0
        col: int = 1
        for ydata in yaxis:
            row += 1.0 / float(cols)
            ydata: Axis = ydata
            fig.append_trace(go.Scatter(
                x=xaxis.data,
                y=ydata.data,
                name=ydata.label,
                line=dict(color=ydata.color, dash=ydata.linestyle)
                # line_color=ydata.color,
                # linestyle=ydata.linestyle
            ), row=ceil(row), col=col)
            col = col + 1 if col < cols else 1
        self.layout(fig, xaxis, yaxis)
        self.show(fig)


import pytest  # important to be able to use @pytest...

from simses.commons.config.data.electrolyzer import ElectrolyzerDataConfig
from simses.commons.constants import Hydrogen
from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.technology.hydrogen.electrolyzer.stack.pem import PemElectrolyzer

"""
Test ideas for Electrolyzers:
- hydrogen Production >=0; 
- hydrogen Production >0 bei power >0
- Pressure anode, cathode >=1


For PEM:
- Pressure anode >= pressure cathode (Differenzdruck über Membran muss >= 0 sein)



"""

'''
#1. Simple Test: Syntax, multiple asserts, import pytest, output in cmd

# def test_propertyof_product():
#     assert demo.product(5,5) == 25 #What should the result be?

#2. Test a Class using pytest's fixtures

# @pytest.fixture(scope="module")#module: gets only one instance
# def make_demo_instance():
#     print("Hey! I got instanciated :)")
#    return Demo()

def test_add(make_demo_instance):
    assert make_demo_instance.add(2, 2) == 4 #quick maths

#3. Parametrize a Test using pytest

@pytest.mark.parametrize('num1, num2, result',
                         [
                             (4, 1, 3),
                             (5, 5, 0)
                         ]
                         )
def test_subtract(num1, num2, result, make_demo_instance):
    print("For debugging reasons it is sometimes nice to print Variables. You can do this with -s", num1)
    assert make_demo_instance.subtract(num1, num2) == result

#4. Practical Example?

#Goto ACDC Example

#5. More Test Types and ressources

#60 Minute Tutorial: https://www.youtube.com/watch?v=bbp_849-RZ4
#Pytest Homepage for examples: https://docs.pytest.org/en/latest/example/index.html
'''


@pytest.fixture(scope="function") #module: gets only one instance
def uut(nominal_power) -> PemElectrolyzer:
    return PemElectrolyzer(nominal_power, ElectrolyzerDataConfig())


def create_electrolyzer_state_from(electrolyzer: PemElectrolyzer) -> ElectrolyzerState:
    state: ElectrolyzerState = ElectrolyzerState(0, 0)
    state.time = 0
    state.power = 0  # W
    state.temperature = 25  # °C
    state.power_loss = 0  # W
    state.hydrogen_production = 0  # mol/s
    state.hydrogen_outflow = 0  # mol/s
    state.oxygen_production = 0  # mol/s
    state.fulfillment = 1.0  # p.u.
    state.soh = 1.0  # p.u.
    state.convection_heat = 0  # W
    state.resistance_increase_cyclic = 0  # p.u.
    state.resistance_increase_calendar = 0  # p.u.
    state.resistance_increase = 0  # p.u.
    state.exchange_current_decrease_cyclic = 1  # p.u.
    state.exchange_current_decrease_calendar = 1  # p.u.
    state.exchange_current_decrease = 1  # p.u.
    state.pressure_anode = Hydrogen.EPS  # barg
    state.pressure_cathode = Hydrogen.EPS  # barg
    state.sat_pressure_h2o = 0.03189409713071401  # bar  h2o saturation pressure at 25°C
    state.part_pressure_h2 = (
                                         1.0 + state.pressure_cathode) - 0.03189409713071401  # bar partial pressure h2 at 25°C and cathode pressure
    state.part_pressure_o2 = (
                                         1.0 + state.pressure_anode) - 0.03189409713071401  # bar partial pressure o2 at 25°C and anode pressure
    state.total_pressure = 20
    state.water_use = 0  # mol/s#
    state.water_outflow_cathode = 0  # mol/s
    state.water_outflow_anode = 0  # mol/s
    state.water_flow = 0  # mol/s
    state.power_water_heating = 0  # W
    state.power_pump = 0  # W
    state.power_gas_drying = 0  # W
    state.power_compressor = 0  # W
    state.total_hydrogen_production = 0  # kg
    state.relative_time = 0  # start
    state.voltage = 1  # stays at 1 so that electrolyzer and fuel cell always see the power indepentently from the voltage a timestep before
    electrolyzer.calculate(state.power, state)
    state.voltage = electrolyzer.get_voltage()
    state.current = electrolyzer.get_current()
    state.current_density = electrolyzer.get_current_density()
    state.max_charge_power = electrolyzer.get_nominal_stack_power()
    state.max_discharge_power = electrolyzer.get_nominal_stack_power()
    return state


@pytest.mark.parametrize('nominal_power, power, temperature, pressure_anode, pressure_cathode, current_result',
                         [
                             (5, 0, 0, 0, 0, 0)
                         ]
                         )
def test_calculate_current(uut: PemElectrolyzer, power, temperature, pressure_anode, pressure_cathode, current_result):
    state: ElectrolyzerState = create_electrolyzer_state_from(uut)
    state.temperature = temperature
    state.pressure_anode = pressure_anode
    state.pressure_cathode = pressure_cathode
    uut.calculate(power, state)
    assert uut.get_current() == current_result


@pytest.mark.parametrize('nominal_power, power, temperature, pressure_anode, pressure_cathode, current_result',
                         [
                             (5, 5, 20, 1, 1, 0)
                         ]
                         )
def test_calculate_geq_current(uut: PemElectrolyzer, power, temperature, pressure_anode, pressure_cathode, current_result):
    state: ElectrolyzerState = create_electrolyzer_state_from(uut)
    state.temperature = temperature
    state.pressure_anode = pressure_anode
    state.pressure_cathode = pressure_cathode
    uut.calculate(power, state)
    assert uut.get_current() > current_result


@pytest.mark.parametrize('nominal_power, power, temperature, pressure_anode, pressure_cathode, voltage_result',
                         [
                             (10, 10, 25, 1, 1, 0),
                             (10, 0, 25, 0, 0, 0)
                         ]
                         )
def test_calculate_voltage(uut: PemElectrolyzer, power, temperature, pressure_anode, pressure_cathode, voltage_result):
    state: ElectrolyzerState = create_electrolyzer_state_from(uut)
    state.temperature = temperature
    state.pressure_anode = pressure_anode
    state.pressure_cathode = pressure_cathode
    uut.calculate(power, state)
    assert uut.get_voltage() > voltage_result


@pytest.mark.parametrize('nominal_power, power, temperature, pressure_anode, pressure_cathode, h2_flow_result',
                         [
                             (10, 10, 25, 1, 1, 0)
                         ]
                         )
def test_calculate_hydrogen_flow_geq(uut: PemElectrolyzer, power, temperature, pressure_anode, pressure_cathode, h2_flow_result):
    state: ElectrolyzerState = create_electrolyzer_state_from(uut)
    state.temperature = temperature
    state.pressure_anode = pressure_anode
    state.pressure_cathode = pressure_cathode
    uut.calculate(power, state)
    assert uut.get_hydrogen_production() > h2_flow_result


@pytest.mark.parametrize('nominal_power, power, temperature, pressure_anode, pressure_cathode, h2_flow_result',
                         [
                             (10, 0, 25, 0, 0, 0)
                         ]
                         )
def test_calculate_hydrogen_flow(uut: PemElectrolyzer, power, temperature, pressure_anode, pressure_cathode, h2_flow_result):
    state: ElectrolyzerState = create_electrolyzer_state_from(uut)
    state.temperature = temperature
    state.pressure_anode = pressure_anode
    state.pressure_cathode = pressure_cathode
    uut.calculate(power, state)
    assert uut.get_hydrogen_production() == h2_flow_result


@pytest.mark.parametrize('nominal_power, power, temperature, pressure_anode, pressure_cathode, get_water_use_result',
                         [
                             (10, 10, 25, 1, 1, 0)
                         ]
                         )
def test_calculate_water_use(uut: PemElectrolyzer, power, temperature, pressure_anode, pressure_cathode, get_water_use_result):
    state: ElectrolyzerState = create_electrolyzer_state_from(uut)
    state.temperature = temperature
    state.pressure_anode = pressure_anode
    state.pressure_cathode = pressure_cathode
    uut.calculate(power, state)
    assert uut.get_water_use() > get_water_use_result



from numpy import sqrt


class RampinelliFit:

    """
    Inverter Equation based on:
    G.A. Rampinelli, A. Krenzinger, and F. Chenlo Romero.
    "Mathematical models for efficiency of inverters used in grid connected photovoltaic systems".
    In: Renewable and Sustainable Energy Reviews 34 (June 2014), pp. 578–587.
    doi: 10.1016/j.rser.2014.03.047.

    efficiency = p / (p + k0 + k1*p k2*p^2), where p = p_ac / p_max.
    """

    def __init__(self, max_power: float, charge_coefficients: tuple, discharge_coefficients: tuple = None):
        self.__max_power: float = max_power
        if discharge_coefficients is None:
            discharge_coefficients = charge_coefficients
        self.__k0_charge, self.__k1_charge, self.__k2_charge = charge_coefficients
        self.__k0_discharge, self.__k1_discharge, self.__k2_discharge = discharge_coefficients

    def get_efficiency_to_ac_reverse(self, dc_power: float) -> float:
        p = (self.__max_power + self.__k1_discharge * self.__max_power) / self.__k2_discharge
        q = (self.__max_power ** 2 * self.__k0_discharge - abs(dc_power) * self.__max_power) / self.__k2_discharge
        return min(0.0, -(-p / 2 + sqrt((p / 2) ** 2 - q)))

    def get_efficiency_to_dc_reverse(self, dc_power: float) -> float:
        p = (abs(dc_power) * self.__max_power * (1 + self.__k1_charge)) / (self.__k2_charge * abs(dc_power) - self.__max_power)
        q = (abs(dc_power) * self.__k0_charge * self.__max_power ** 2) / (self.__k2_charge * abs(dc_power) - self.__max_power)
        return max(0.0, -p / 2 + sqrt((p / 2) ** 2 - q))

    def get_efficiency_to_ac(self, power: float) -> float:
        power_ratio = abs(power) / self.__max_power
        if power_ratio < 0.0 or power_ratio > 1.0:
            raise Exception('Power ratio is not possible: ' + str(power_ratio))
        return power_ratio / (power_ratio + self.__k0_discharge + self.__k1_discharge * power_ratio +
                              self.__k2_discharge * power_ratio ** 2)

    def get_efficiency_to_dc(self, power: float) -> float:
        power_ratio = abs(power) / self.__max_power
        if power_ratio < 0.0 or power_ratio > 1.0:
            raise Exception('Power ratio is not possible: ' + str(power_ratio))
        return power_ratio / (power_ratio + self.__k0_charge + self.__k1_charge * power_ratio +
                              self.__k2_charge * power_ratio ** 2)



from abc import ABC, abstractmethod


class PressureController(ABC):
    """ This controller controls the pressure of anode and cathode side of the electrolyzer like a control valve"""
    def __init__(self):
        super().__init__()

    @abstractmethod
    def calculate_n_h2_out(self, pressure_cathode: float, n_h2_prod: float, timestep: float, pressure_factor: float) -> float:
        pass

    @abstractmethod
    def calculate_n_o2_out(self, pressure_anode: float, n_o2_prod: float, timestep: float) -> float:
        pass

    @abstractmethod
    def calculate_n_h2_in(self, pressure_anode: float, n_h2_used: float) -> float:
        pass

    @abstractmethod
    def calculate_n_o2_in(self, pressure_cathode: float, n_o2_used: float) -> float:
        pass
